"""

    Data types and validators (PointProperty) for working with neo4j's spatial data types through neomodel.

    `spatial_properties` offers two classes: NeomodelPoint, PointProperty that marshal data to and from a neo4j dbms,
    with the added capability of being Shapely objects. Therefore, points retrieved with Neomodel can readily be used
    in further geometric (via Shapely) or geospatial (via PySAL) operations.

    * More information on Neo4J's spatial data types:
        * https://neo4j.com/docs/developer-manual/3.4/cypher/syntax/spatial/

    * More information on the Python driver's data types:
        * https://neo4j.com/docs/api/python-driver/1.7-preview/types.html#spatial-types

    * More information about Shapely's spatial data types:
        * http://toblerity.org/shapely/manual.html#geometric-objects

    * More information about PySAL through Shapely:
        * https://pysal.readthedocs.io/en/latest/users/tutorials/shapely.html
"""

__author__ = "Athanasios Anastasiou"

import neo4j.spatial

# If shapely is not installed, its import will fail and the spatial properties will not be available
try:
    from shapely import __version__ as shapely_version
    from shapely.coords import CoordinateSequence
    from shapely.geometry import Point as ShapelyPoint
except ImportError as exc:
    raise ImportError(
        "NEOMODEL ERROR: Shapely not found. If required, you can install Shapely via "
        "`pip install shapely`."
    ) from exc

from neomodel.properties import Property, validator

# Note: Depending on how Neo4J decides to handle the resolution of geographical points, these two
# private attributes might have to be updated in the future or removed altogether.
# They help in trying to resolve which class needs to be instantiated based on the data received from the dbms
# and doing it in a concise way that minimises errors.
#
# Acceptable Coordinate Reference Systems
ACCEPTABLE_CRS = ["cartesian", "cartesian-3d", "wgs-84", "wgs-84-3d"]
# A simple CRS to SRID mapping
SRID_TO_CRS = {
    7203: "cartesian",
    9157: "cartesian-3d",
    4326: "wgs-84",
    4979: "wgs-84-3d",
}

# Taking into account the Shapely 2.0.0 changes in the way POINT objects are initialisd.
if int("".join(shapely_version.split(".")[0:3])) < 200:

    class NeomodelPoint(ShapelyPoint):
        """
        Abstracts the Point spatial data type of Neo4j.

        Note:
        At the time of writing, Neo4j supports 2 main variants of Point:
            1. A generic point defined over a Cartesian plane
                * The minimum data to define a point is x, y [,z] when crs is either "cartesian" or "cartesian-3d"
            2. A generic point defined over the WGS84 ellipsoid
                * The minimum data to define a point is longitude, latitude [,Height] and the crs is then assumed
                  to be "wgs-84".
        """

        # def __init__(self, *args, crs=None, x=None, y=None, z=None, latitude=None, longitude=None, height=None, **kwargs):
        def __init__(self, *args, **kwargs):
            """
            Creates a NeomodelPoint.

            :param args: Positional arguments to emulate the behaviour of Shapely's Point (and specifically the copy
            constructor)
            :type args: list
            :param crs: Coordinate Reference System, must be one of ['cartesian', 'cartesian-3d', 'wgs-84', 'wgs-84-3d']
            :type crs: str
            :param x: x coordinate of point
            :type x: float
            :param y: y coordinate of point
            :type y: float
            :param z: z coordinate of point if the crs is cartesian-3d
            :type z: float
            :param latitude: Latitude of point
            :type latitude: float
            :param longitude: Longitude of point
            :type longitude: float
            :param height: Height of point if the crs is wgs-84-3d
            :type height: float
            :param kwargs: Dictionary of keyword arguments
            :type kwargs: dict
            """

            # Python2.7 Workaround for the order that the arguments get passed to the functions
            crs = kwargs.pop("crs", None)
            x = kwargs.pop("x", None)
            y = kwargs.pop("y", None)
            z = kwargs.pop("z", None)
            longitude = kwargs.pop("longitude", None)
            latitude = kwargs.pop("latitude", None)
            height = kwargs.pop("height", None)

            _x, _y, _z = None, None, None

            # CRS validity check is common to both types of constructors that follow
            if crs is not None and crs not in ACCEPTABLE_CRS:
                raise ValueError(
                    f"Invalid CRS({crs}). Expected one of {','.join(ACCEPTABLE_CRS)}"
                )
            self._crs = crs

            # If positional arguments have been supplied, then this is a possible call to the copy constructor or
            # initialisation by a coordinate iterable as per ShapelyPoint constructor.
            if len(args) > 0:
                # If a coordinate iterable was passed, emulate a call with x,y[,z] parameters
                if isinstance(args[0], (tuple, list)):
                    # Check dimensionality of tuple
                    if len(args[0]) < 2 or len(args[0]) > 3:
                        raise ValueError(
                            f"Invalid vector dimensions. Expected 2 or 3, received {len(args[0])}"
                        )
                    x = args[0][0]
                    y = args[0][1]
                    if len(args[0]) == 3:
                        z = args[0][2]
                # If another "Point" was passed, then this is a call to the copy constructor
                elif isinstance(args[0], ShapelyPoint):
                    super().__init__(args[0])
                    # If the other Point was a NeomodelPoint then it bears the CRS that is used to
                    # interpret the points and this has to be carried over.
                    if isinstance(args[0], NeomodelPoint):
                        self._crs = args[0]._crs
                    else:
                        # This allows NeomodelPoint((0,0),crs="wgs-84") which will interpret the tuple as
                        # (longitude,latitude) even though it was not specified as such with the named arguments.
                        #
                        # NOTE: Notice the indexing on the coordinates `args[0].coords[0]`. Coordinates are always an
                        # iterable which for Points has length of 1 but for other geometrical objects (e.g. boundaries)
                        # might be longer. This will come back to bite you, if you assume that `coords` is what it pretends
                        # to be (i.e. coords gives access to the actual coordinates NOT an iterable).
                        #
                        if len(args[0].coords[0]) == 2:
                            if crs is None:
                                self._crs = "cartesian"
                        elif len(args[0].coords[0]) == 3:
                            if crs is None:
                                self._crs = "cartesian-3d"
                        else:
                            raise ValueError(
                                f"Invalid vector dimensions. Expected 2 or 3, received {len(args[0].coords[0])}"
                            )
                    return
                else:
                    raise TypeError(
                        f"Invalid object passed to copy constructor. Expected NeomodelPoint or shapely Point, received {type(args[0])}"
                    )

            # Initialisation is either via x,y[,z] XOR longitude,latitude[,height]. Specifying both leads to an error.
            if any(i is not None for i in [x, y, z]) and any(
                i is not None for i in [latitude, longitude, height]
            ):
                raise ValueError(
                    "Invalid instantiation via arguments. "
                    "A Point can be defined either by x,y,z coordinates OR latitude,longitude,height but not "
                    "a combination of these terms"
                )

            # Specifying no initialisation argument at this point in the constructor is flagged as an error
            if all(i is None for i in [x, y, z, latitude, longitude, height]):
                raise ValueError(
                    "Invalid instantiation via no arguments. "
                    "A Point needs default values either in x,y,z or longitude, latitude, height coordinates"
                )

            # Geographical Point Initialisation
            if latitude is not None and longitude is not None:
                if height is not None:
                    if self._crs is None:
                        self._crs = "wgs-84-3d"
                    _z = height
                else:
                    if self._crs is None:
                        self._crs = "wgs-84"
                _x = longitude
                _y = latitude

            # Geometrical Point Initialisation
            if x is not None and y is not None:
                if z is not None:
                    if self._crs is None:
                        self._crs = "cartesian-3d"
                    _z = z
                else:
                    if self._crs is None:
                        self._crs = "cartesian"
                _x = x
                _y = y

            if _z is None:
                if "-3d" not in self._crs:
                    super().__init__((float(_x), float(_y)), **kwargs)
                else:
                    raise ValueError(
                        f"Invalid vector dimensions(2) for given CRS({self._crs})."
                    )
            else:
                if "-3d" in self._crs:
                    super().__init__((float(_x), float(_y), float(_z)), **kwargs)
                else:
                    raise ValueError(
                        f"Invalid vector dimensions(3) for given CRS({self._crs})."
                    )

        @property
        def crs(self):
            return self._crs

        @property
        def x(self):
            if not self._crs.startswith("cartesian"):
                raise TypeError(
                    f'Invalid coordinate ("x") for points defined over {self.crs}'
                )
            return super().x

        @property
        def y(self):
            if not self._crs.startswith("cartesian"):
                raise TypeError(
                    f'Invalid coordinate ("y") for points defined over {self.crs}'
                )
            return super().y

        @property
        def z(self):
            if self._crs != "cartesian-3d":
                raise TypeError(
                    f'Invalid coordinate ("z") for points defined over {self.crs}'
                )
            return super().z

        @property
        def latitude(self):
            if not self._crs.startswith("wgs-84"):
                raise TypeError(
                    f'Invalid coordinate ("latitude") for points defined over {self.crs}'
                )
            return super().y

        @property
        def longitude(self):
            if not self._crs.startswith("wgs-84"):
                raise TypeError(
                    f'Invalid coordinate ("longitude") for points defined over {self.crs}'
                )
            return super().x

        @property
        def height(self):
            if self._crs != "wgs-84-3d":
                raise TypeError(
                    f'Invalid coordinate ("height") for points defined over {self.crs}'
                )
            return super().z

        # The following operations are necessary here due to the way queries (and more importantly their parameters) get
        # combined and evaluated in neomodel. Specifically, query expressions get duplicated with deep copies and any valid
        # datatype values should also implement these operations.
        def __copy__(self):
            return NeomodelPoint(self)

        def __deepcopy__(self, memo):
            return NeomodelPoint(self)

else:

    class NeomodelPoint:
        """
        Abstracts the Point spatial data type of Neo4j.

        Note:
        At the time of writing, Neo4j supports 2 main variants of Point:
            1. A generic point defined over a Cartesian plane
                * The minimum data to define a point is x, y [,z] when crs is either "cartesian" or "cartesian-3d"
            2. A generic point defined over the WGS84 ellipsoid
                * The minimum data to define a point is longitude, latitude [,Height] and the crs is then assumed
                  to be "wgs-84".
        """

        # def __init__(self, *args, crs=None, x=None, y=None, z=None, latitude=None, longitude=None, height=None, **kwargs):
        def __init__(self, *args, **kwargs):
            """
            Creates a NeomodelPoint.

            :param args: Positional arguments to emulate the behaviour of Shapely's Point (and specifically the copy
            constructor)
            :type args: list
            :param crs: Coordinate Reference System, must be one of ['cartesian', 'cartesian-3d', 'wgs-84', 'wgs-84-3d']
            :type crs: str
            :param x: x coordinate of point
            :type x: float
            :param y: y coordinate of point
            :type y: float
            :param z: z coordinate of point if the crs is cartesian-3d
            :type z: float
            :param latitude: Latitude of point
            :type latitude: float
            :param longitude: Longitude of point
            :type longitude: float
            :param height: Height of point if the crs is wgs-84-3d
            :type height: float
            :param kwargs: Dictionary of keyword arguments
            :type kwargs: dict
            """

            # Python2.7 Workaround for the order that the arguments get passed to the functions
            crs = kwargs.pop("crs", None)
            x = kwargs.pop("x", None)
            y = kwargs.pop("y", None)
            z = kwargs.pop("z", None)
            longitude = kwargs.pop("longitude", None)
            latitude = kwargs.pop("latitude", None)
            height = kwargs.pop("height", None)

            _x, _y, _z = None, None, None

            # CRS validity check is common to both types of constructors that follow
            if crs is not None and crs not in ACCEPTABLE_CRS:
                raise ValueError(
                    f"Invalid CRS({crs}). Expected one of {','.join(ACCEPTABLE_CRS)}"
                )
            self._crs = crs

            # If positional arguments have been supplied, then this is a possible call to the copy constructor or
            # initialisation by a coordinate iterable as per ShapelyPoint constructor.
            if len(args) > 0:
                # If a coordinate iterable was passed, emulate a call with x,y[,z] parameters
                if isinstance(args[0], (tuple, list)):
                    # Check dimensionality of tuple
                    if len(args[0]) < 2 or len(args[0]) > 3:
                        raise ValueError(
                            f"Invalid vector dimensions. Expected 2 or 3, received {len(args[0])}"
                        )
                    x = args[0][0]
                    y = args[0][1]
                    if len(args[0]) == 3:
                        z = args[0][2]
                # If another "Point" was passed, then this is a call to the copy constructor
                elif isinstance(args[0], (ShapelyPoint, NeomodelPoint)):
                    if isinstance(args[0], ShapelyPoint):
                        self._shapely_point = ShapelyPoint(args[0])
                    # If the other Point was a NeomodelPoint then it bears the CRS that is used to
                    # interpret the points and this has to be carried over.
                    if isinstance(args[0], NeomodelPoint):
                        self._shapely_point = ShapelyPoint(args[0].coords)
                        self._crs = args[0].crs
                    else:
                        # This allows NeomodelPoint((0,0),crs="wgs-84") which will interpret the tuple as
                        # (longitude,latitude) even though it was not specified as such with the named arguments.
                        #
                        # NOTE: Notice the indexing on the coordinates `args[0].coords[0]`. Coordinates are always an
                        # iterable which for Points has length of 1 but for other geometrical objects (e.g. boundaries)
                        # might be longer. This will come back to bite you, if you assume that `coords` is what it pretends
                        # to be (i.e. coords gives access to the actual coordinates NOT an iterable).
                        #
                        if len(args[0].coords[0]) == 2:
                            if crs is None:
                                self._crs = "cartesian"
                        elif len(args[0].coords[0]) == 3:
                            if crs is None:
                                self._crs = "cartesian-3d"
                        else:
                            raise ValueError(
                                f"Invalid vector dimensions. Expected 2 or 3, received {len(args[0].coords[0])}"
                            )
                    return
                else:
                    raise TypeError(
                        f"Invalid object passed to copy constructor. Expected NeomodelPoint or shapely Point, received {type(args[0])}"
                    )

            # Initialisation is either via x,y[,z] XOR longitude,latitude[,height]. Specifying both leads to an error.
            if any(i is not None for i in [x, y, z]) and any(
                i is not None for i in [latitude, longitude, height]
            ):
                raise ValueError(
                    "Invalid instantiation via arguments. "
                    "A Point can be defined either by x,y,z coordinates OR latitude,longitude,height but not "
                    "a combination of these terms"
                )

            # Specifying no initialisation argument at this point in the constructor is flagged as an error
            if all(i is None for i in [x, y, z, latitude, longitude, height]):
                raise ValueError(
                    "Invalid instantiation via no arguments. "
                    "A Point needs default values either in x,y,z or longitude, latitude, height coordinates"
                )

            # Geographical Point Initialisation
            if latitude is not None and longitude is not None:
                if height is not None:
                    if self._crs is None:
                        self._crs = "wgs-84-3d"
                    _z = height
                else:
                    if self._crs is None:
                        self._crs = "wgs-84"
                _x = longitude
                _y = latitude

            # Geometrical Point Initialisation
            if x is not None and y is not None:
                if z is not None:
                    if self._crs is None:
                        self._crs = "cartesian-3d"
                    _z = z
                else:
                    if self._crs is None:
                        self._crs = "cartesian"
                _x = x
                _y = y

            if _z is None:
                if "-3d" not in self._crs:
                    self._shapely_point = ShapelyPoint((float(_x), float(_y)))
                else:
                    raise ValueError(
                        f"Invalid vector dimensions(2) for given CRS({self._crs})."
                    )
            else:
                if "-3d" in self._crs:
                    self._shapely_point = ShapelyPoint(
                        (float(_x), float(_y), float(_z))
                    )
                else:
                    raise ValueError(
                        f"Invalid vector dimensions(3) for given CRS({self._crs})."
                    )

        @property
        def crs(self):
            return self._crs

        @property
        def x(self):
            if not self._crs.startswith("cartesian"):
                raise TypeError(
                    f'Invalid coordinate ("x") for points defined over {self.crs}'
                )
            return self._shapely_point.x

        @property
        def y(self):
            if not self._crs.startswith("cartesian"):
                raise TypeError(
                    f'Invalid coordinate ("y") for points defined over {self.crs}'
                )
            return self._shapely_point.y

        @property
        def z(self):
            if self._crs != "cartesian-3d":
                raise TypeError(
                    f'Invalid coordinate ("z") for points defined over {self.crs}'
                )
            return self._shapely_point.z

        @property
        def latitude(self):
            if not self._crs.startswith("wgs-84"):
                raise TypeError(
                    f'Invalid coordinate ("latitude") for points defined over {self.crs}'
                )
            return self._shapely_point.y

        @property
        def longitude(self):
            if not self._crs.startswith("wgs-84"):
                raise TypeError(
                    f'Invalid coordinate ("longitude") for points defined over {self.crs}'
                )
            return self._shapely_point.x

        @property
        def height(self):
            if self._crs != "wgs-84-3d":
                raise TypeError(
                    f'Invalid coordinate ("height") for points defined over {self.crs}'
                )
            return self._shapely_point.z

        # The following operations are necessary here due to the way queries (and more importantly their parameters) get
        # combined and evaluated in neomodel. Specifically, query expressions get duplicated with deep copies and any valid
        # datatype values should also implement these operations.
        def __copy__(self):
            return NeomodelPoint(self)

        def __deepcopy__(self, memo):
            return NeomodelPoint(self)

        def __getattr__(self, attr):
            """
            Route messages to the right underlying object.
            """
            return getattr(self._shapely_point, attr)

        def __eq__(self, other):
            """
            Compare objects by value
            """
            if not isinstance(other, (ShapelyPoint, NeomodelPoint)):
                raise ValueError(
                    f"NeomodelPoint equality comparison expected NeomodelPoint or Shapely Point, received {type(other)}"
                )
            else:
                if isinstance(other, ShapelyPoint):
                    return self.coords[0] == other.coords[0]
                else:
                    return self.coords[0] == other.coords[0] and self.crs == other.crs


class PointProperty(Property):
    """
    Validates points which can participate in spatial queries.
    """

    form_field_class = "PointField"
    # The CRS that this property is expected to be expressed in.
    _crs = None

    def __init__(self, *args, **kwargs):
        """
        A Point property that requires at least its CRS to be known to offer proper validation.

        :param crs: Coordinate Reference System
        :type crs: str
        :param kwargs: Dictionary of arguments
        :type kwargs: dict
        """
        if "crs" in kwargs:
            crs = kwargs["crs"]
            del kwargs["crs"]
        else:
            crs = None

        if crs is None or (crs not in ACCEPTABLE_CRS):
            raise ValueError(
                f"Invalid CRS({crs}). Point properties require CRS to be one of {','.join(ACCEPTABLE_CRS)}"
            )

        # If a default value is passed and it is not a callable, then make sure it is in the right type
        if (
            "default" in kwargs
            and not hasattr(kwargs["default"], "__call__")
            and not isinstance(kwargs["default"], NeomodelPoint)
        ):
            raise TypeError(
                f"Invalid default value. Expected NeomodelPoint, received {type(kwargs['default'])}"
            )

        super().__init__(*args, **kwargs)
        self._crs = crs

    @validator
    def inflate(self, value):
        """
        Handles the marshalling from Neo4J POINT to NeomodelPoint

        :param value: Value returned from the database
        :type value: Neo4J POINT
        :return: NeomodelPoint
        """
        if not isinstance(value, neo4j.spatial.Point):
            raise TypeError(
                f"Invalid datatype to inflate. Expected POINT datatype, received {type(value)}"
            )

        try:
            value_point_crs = SRID_TO_CRS[value.srid]
        except KeyError as e:
            raise ValueError(
                f"Invalid SRID to inflate. Expected one of {SRID_TO_CRS.keys()}, received {value.srid}"
            ) from e

        if self._crs != value_point_crs:
            raise ValueError(
                f"Invalid CRS. Expected POINT defined over {self._crs}, received {value_point_crs}"
            )
        # cartesian
        if value.srid == 7203:
            return NeomodelPoint(x=value.x, y=value.y)
        # cartesian-3d
        if value.srid == 9157:
            return NeomodelPoint(x=value.x, y=value.y, z=value.z)
        # wgs-84
        if value.srid == 4326:
            return NeomodelPoint(longitude=value.longitude, latitude=value.latitude)
        # wgs-83-3d
        if value.srid == 4979:
            return NeomodelPoint(
                longitude=value.longitude,
                latitude=value.latitude,
                height=value.height,
            )
        return None

    @validator
    def deflate(self, value):
        """
        Handles the marshalling from NeomodelPoint to Neo4J POINT

        :param value: The point that was assigned as value to a property in the model
        :type value: NeomodelPoint
        :return: Neo4J POINT
        """
        if not isinstance(value, NeomodelPoint):
            raise TypeError(
                f"Invalid datatype to deflate. Expected NeomodelPoint, received {type(value)}"
            )

        if value.crs != self._crs:
            raise ValueError(
                f"Invalid CRS. Expected NeomodelPoint defined over {self._crs}, received NeomodelPoint defined over {value.crs}"
            )

        if value.crs == "cartesian-3d":
            return neo4j.spatial.CartesianPoint((value.x, value.y, value.z))
        if value.crs == "cartesian":
            return neo4j.spatial.CartesianPoint((value.x, value.y))
        if value.crs == "wgs-84":
            return neo4j.spatial.WGS84Point((value.longitude, value.latitude))
        if value.crs == "wgs-84-3d":
            return neo4j.spatial.WGS84Point(
                (value.longitude, value.latitude, value.height)
            )
        return None



from test._async_compat import mark_async_test

from pytest import raises

from neomodel import (
    AsyncRelationshipFrom,
    AsyncRelationshipTo,
    AsyncStructuredNode,
    IntegerProperty,
    StringProperty,
    UniqueIdProperty,
    config,
)
from neomodel._async_compat.util import AsyncUtil
from neomodel.exceptions import DeflateError, UniqueProperty

config.AUTO_INSTALL_LABELS = True


class UniqueUser(AsyncStructuredNode):
    uid = UniqueIdProperty()
    name = StringProperty()
    age = IntegerProperty()


@mark_async_test
async def test_unique_id_property_batch():
    users = await UniqueUser.create(
        {"name": "bob", "age": 2}, {"name": "ben", "age": 3}
    )

    assert users[0].uid != users[1].uid

    users = await UniqueUser.get_or_create(
        {"uid": users[0].uid}, {"name": "bill", "age": 4}
    )

    assert users[0].name == "bob"
    assert users[1].uid


class Customer(AsyncStructuredNode):
    email = StringProperty(unique_index=True, required=True)
    age = IntegerProperty(index=True)


@mark_async_test
async def test_batch_create():
    users = await Customer.create(
        {"email": "jim1@aol.com", "age": 11},
        {"email": "jim2@aol.com", "age": 7},
        {"email": "jim3@aol.com", "age": 9},
        {"email": "jim4@aol.com", "age": 7},
        {"email": "jim5@aol.com", "age": 99},
    )
    assert len(users) == 5
    assert users[0].age == 11
    assert users[1].age == 7
    assert users[1].email == "jim2@aol.com"
    assert await Customer.nodes.get(email="jim1@aol.com")


@mark_async_test
async def test_batch_create_or_update():
    users = await Customer.create_or_update(
        {"email": "merge1@aol.com", "age": 11},
        {"email": "merge2@aol.com"},
        {"email": "merge3@aol.com", "age": 1},
        {"email": "merge2@aol.com", "age": 2},
    )
    assert len(users) == 4
    assert users[1] == users[3]
    merge_1: Customer = await Customer.nodes.get(email="merge1@aol.com")
    assert merge_1.age == 11

    more_users = await Customer.create_or_update(
        {"email": "merge1@aol.com", "age": 22},
        {"email": "merge4@aol.com", "age": None},
    )
    assert len(more_users) == 2
    assert users[0] == more_users[0]
    merge_1 = await Customer.nodes.get(email="merge1@aol.com")
    assert merge_1.age == 22


@mark_async_test
async def test_batch_validation():
    # test validation in batch create
    with raises(DeflateError):
        await Customer.create(
            {"email": "jim1@aol.com", "age": "x"},
        )


@mark_async_test
async def test_batch_index_violation():
    for u in await Customer.nodes:
        await u.delete()

    users = await Customer.create(
        {"email": "jim6@aol.com", "age": 3},
    )
    assert users
    with raises(UniqueProperty):
        await Customer.create(
            {"email": "jim6@aol.com", "age": 3},
            {"email": "jim7@aol.com", "age": 5},
        )

    # not found
    if AsyncUtil.is_async_code:
        assert not await Customer.nodes.filter(email="jim7@aol.com").check_bool()
    else:
        assert not Customer.nodes.filter(email="jim7@aol.com")


class Dog(AsyncStructuredNode):
    name = StringProperty(required=True)
    owner = AsyncRelationshipTo("Person", "owner")


class Person(AsyncStructuredNode):
    name = StringProperty(unique_index=True)
    pets = AsyncRelationshipFrom("Dog", "owner")


@mark_async_test
async def test_get_or_create_with_rel():
    create_bob = await Person.get_or_create({"name": "Bob"})
    bob = create_bob[0]
    bobs_gizmo = await Dog.get_or_create({"name": "Gizmo"}, relationship=bob.pets)

    create_tim = await Person.get_or_create({"name": "Tim"})
    tim = create_tim[0]
    tims_gizmo = await Dog.get_or_create({"name": "Gizmo"}, relationship=tim.pets)

    # not the same gizmo
    assert bobs_gizmo[0] != tims_gizmo[0]



import io
from test._async_compat import mark_async_test
from unittest.mock import patch

from pytest import raises

from neomodel import (
    AsyncOne,
    AsyncOneOrMore,
    AsyncRelationshipFrom,
    AsyncRelationshipTo,
    AsyncStructuredNode,
    AsyncZeroOrMore,
    AsyncZeroOrOne,
    AttemptedCardinalityViolation,
    CardinalityViolation,
    IntegerProperty,
    StringProperty,
    adb,
    config,
)


class HairDryer(AsyncStructuredNode):
    version = IntegerProperty()


class ScrewDriver(AsyncStructuredNode):
    version = IntegerProperty()


class Car(AsyncStructuredNode):
    version = IntegerProperty()


class Monkey(AsyncStructuredNode):
    name = StringProperty()
    dryers = AsyncRelationshipTo("HairDryer", "OWNS_DRYER", cardinality=AsyncZeroOrMore)
    driver = AsyncRelationshipTo(
        "ScrewDriver", "HAS_SCREWDRIVER", cardinality=AsyncZeroOrOne
    )
    car = AsyncRelationshipTo("Car", "HAS_CAR", cardinality=AsyncOneOrMore)
    toothbrush = AsyncRelationshipTo(
        "ToothBrush", "HAS_TOOTHBRUSH", cardinality=AsyncOne
    )


class ToothBrush(AsyncStructuredNode):
    name = StringProperty()


class Owner(AsyncStructuredNode):
    name = StringProperty(required=True)
    pets = AsyncRelationshipTo("Pet", "OWNS")


class Pet(AsyncStructuredNode):
    name = StringProperty(required=True)
    owner = AsyncRelationshipFrom("Owner", "OWNS", cardinality=AsyncOne)


class Company(AsyncStructuredNode):
    name = StringProperty(required=True)
    employees = AsyncRelationshipTo("Employee", "EMPLOYS")


class Employee(AsyncStructuredNode):
    name = StringProperty(required=True)
    employer = AsyncRelationshipFrom("Company", "EMPLOYS", cardinality=AsyncZeroOrOne)


class Manager(AsyncStructuredNode):
    name = StringProperty(required=True)
    assistant = AsyncRelationshipTo("Assistant", "MANAGES", cardinality=AsyncOne)


class Assistant(AsyncStructuredNode):
    name = StringProperty(required=True)
    boss = AsyncRelationshipFrom("Manager", "MANAGES", cardinality=AsyncOne)


@mark_async_test
async def test_cardinality_zero_or_more():
    m = await Monkey(name="tim").save()
    assert await m.dryers.all() == []
    single_dryer = await m.driver.single()
    assert single_dryer is None
    h = await HairDryer(version=1).save()

    await m.dryers.connect(h)
    assert len(await m.dryers.all()) == 1
    single_dryer = await m.dryers.single()
    assert single_dryer.version == 1

    await m.dryers.disconnect(h)
    assert await m.dryers.all() == []
    single_dryer = await m.driver.single()
    assert single_dryer is None

    h2 = await HairDryer(version=2).save()
    await m.dryers.connect(h)
    await m.dryers.connect(h2)
    await m.dryers.disconnect_all()
    assert await m.dryers.all() == []
    single_dryer = await m.driver.single()
    assert single_dryer is None


@mark_async_test
async def test_cardinality_zero_or_one():
    m = await Monkey(name="bob").save()
    assert await m.driver.all() == []
    single_driver = await m.driver.single()
    assert await m.driver.single() is None
    h = await ScrewDriver(version=1).save()

    await m.driver.connect(h)
    assert len(await m.driver.all()) == 1
    single_driver = await m.driver.single()
    assert single_driver.version == 1

    j = await ScrewDriver(version=2).save()
    with raises(AttemptedCardinalityViolation):
        await m.driver.connect(j)

    await m.driver.reconnect(h, j)
    single_driver = await m.driver.single()
    assert single_driver.version == 2

    # Forcing creation of a second ToothBrush to go around
    # AttemptedCardinalityViolation
    await adb.cypher_query(
        """
        MATCH (m:Monkey WHERE m.name="bob")
        CREATE (s:ScrewDriver {version:3})
        WITH m, s
        CREATE (m)-[:HAS_SCREWDRIVER]->(s)
    """
    )
    with raises(
        CardinalityViolation, match=r"CardinalityViolation: Expected: .*, got: 2."
    ):
        await m.driver.all()


@mark_async_test
async def test_cardinality_one_or_more():
    m = await Monkey(name="jerry").save()

    with raises(CardinalityViolation):
        await m.car.all()

    with raises(CardinalityViolation):
        await m.car.single()

    c = await Car(version=2).save()
    await m.car.connect(c)
    single_car = await m.car.single()
    assert single_car.version == 2

    cars = await m.car.all()
    assert len(cars) == 1

    with raises(AttemptedCardinalityViolation):
        await m.car.disconnect(c)

    d = await Car(version=3).save()
    await m.car.connect(d)
    cars = await m.car.all()
    assert len(cars) == 2

    await m.car.disconnect(d)
    cars = await m.car.all()
    assert len(cars) == 1


@mark_async_test
async def test_cardinality_one():
    m = await Monkey(name="jerry").save()

    with raises(
        CardinalityViolation, match=r"CardinalityViolation: Expected: .*, got: none."
    ):
        await m.toothbrush.all()

    with raises(CardinalityViolation):
        await m.toothbrush.single()

    b = await ToothBrush(name="Jim").save()
    await m.toothbrush.connect(b)
    single_toothbrush = await m.toothbrush.single()
    assert single_toothbrush.name == "Jim"

    x = await ToothBrush(name="Jim").save()
    with raises(AttemptedCardinalityViolation):
        await m.toothbrush.connect(x)

    with raises(AttemptedCardinalityViolation):
        await m.toothbrush.disconnect(b)

    with raises(AttemptedCardinalityViolation):
        await m.toothbrush.disconnect_all()

    # Forcing creation of a second ToothBrush to go around
    # AttemptedCardinalityViolation
    await adb.cypher_query(
        """
        MATCH (m:Monkey WHERE m.name="jerry")
        CREATE (t:ToothBrush {name:"Jim"})
        WITH m, t
        CREATE (m)-[:HAS_TOOTHBRUSH]->(t)
    """
    )
    with raises(
        CardinalityViolation, match=r"CardinalityViolation: Expected: .*, got: 2."
    ):
        await m.toothbrush.all()

    jp = Monkey(name="Jean-Pierre")
    with raises(ValueError, match="Node has not been saved cannot connect!"):
        await jp.toothbrush.connect(b)


@mark_async_test
async def test_relationship_from_one_cardinality_enforced():
    """
    Test that RelationshipFrom with cardinality=One prevents multiple connections.

    This addresses the GitHub issue where RelationshipFrom cardinality constraints
    were not being enforced.
    """
    # Setup
    config.SOFT_INVERSE_CARDINALITY_CHECK = False
    owner1 = await Owner(name="Alice").save()
    owner2 = await Owner(name="Bob").save()
    pet = await Pet(name="Fluffy").save()

    # First connection should succeed
    await owner1.pets.connect(pet)

    # Verify connection was established
    assert await pet.owner.single() == owner1
    assert pet in await owner1.pets.all()

    # Second connection should fail due to RelationshipFrom cardinality=One
    with raises(AttemptedCardinalityViolation):
        await owner2.pets.connect(pet)

    stream = io.StringIO()
    with patch("sys.stdout", new=stream):
        config.SOFT_INVERSE_CARDINALITY_CHECK = True
        await owner2.pets.connect(pet)
        assert pet in await owner2.pets.all()

    console_output = stream.getvalue()
    assert "Cardinality violation detected" in console_output
    assert "Soft check is enabled so the relationship will be created" in console_output
    assert "strict check will be enabled by default in version 6.0" in console_output


@mark_async_test
async def test_relationship_from_zero_or_one_cardinality_enforced():
    """
    Test that RelationshipFrom with cardinality=ZeroOrOne prevents multiple connections.
    """
    # Setup
    config.SOFT_INVERSE_CARDINALITY_CHECK = False
    company1 = await Company(name="TechCorp").save()
    company2 = await Company(name="StartupInc").save()
    employee = await Employee(name="John").save()

    # First connection should succeed
    await company1.employees.connect(employee)

    # Verify connection was established
    assert await employee.employer.single() == company1
    assert employee in await company1.employees.all()

    # Second connection should fail due to RelationshipFrom cardinality=ZeroOrOne
    with raises(AttemptedCardinalityViolation):
        await company2.employees.connect(employee)

    stream = io.StringIO()
    with patch("sys.stdout", new=stream):
        config.SOFT_INVERSE_CARDINALITY_CHECK = True
        await company2.employees.connect(employee)
        assert employee in await company2.employees.all()

    console_output = stream.getvalue()
    assert "Cardinality violation detected" in console_output
    assert "Soft check is enabled so the relationship will be created" in console_output
    assert "strict check will be enabled by default in version 6.0" in console_output


@mark_async_test
async def test_bidirectional_cardinality_validation():
    """
    Test that cardinality is validated on both ends when both sides have constraints.
    """
    # Setup
    config.SOFT_INVERSE_CARDINALITY_CHECK = False
    manager1 = await Manager(name="Sarah").save()
    manager2 = await Manager(name="David").save()
    assistant = await Assistant(name="Alex").save()

    # First connection should succeed
    await manager1.assistant.connect(assistant)

    # Verify bidirectional connection
    assert await manager1.assistant.single() == assistant
    assert await assistant.boss.single() == manager1

    # Second manager trying to connect to same assistant should fail
    with raises(AttemptedCardinalityViolation):
        await manager2.assistant.connect(assistant)

    stream = io.StringIO()
    with patch("sys.stdout", new=stream):
        config.SOFT_INVERSE_CARDINALITY_CHECK = True
        await manager2.assistant.connect(assistant)
        assert assistant in await manager2.assistant.all()

    console_output = stream.getvalue()
    assert "Cardinality violation detected" in console_output
    assert "Soft check is enabled so the relationship will be created" in console_output
    assert "strict check will be enabled by default in version 6.0" in console_output



from test._async_compat import mark_sync_test
from test.sync_.test_relationships import Country

from neomodel import RelationshipTo, StringProperty, StructuredNode


class Cat(StructuredNode):
    name = StringProperty()
    # Relationship is defined using a relative class path
    is_from = RelationshipTo(".test_relationships.Country", "IS_FROM")


@mark_sync_test
def test_relative_relationship():
    a = Cat(name="snufkin").save()
    assert a

    c = Country(code="MG").save()
    assert c

    # connecting an instance of the class defined above
    # the next statement will fail if there's a type mismatch
    a.is_from.connect(c)
    assert a.is_from.is_connected(c)



from test._async_compat import mark_async_test

import pytest
from neo4j.api import Bookmarks
from neo4j.exceptions import ClientError, TransactionError
from pytest import raises

from neomodel import AsyncStructuredNode, StringProperty, UniqueProperty, adb


class APerson(AsyncStructuredNode):
    name = StringProperty(unique_index=True)


@mark_async_test
async def test_rollback_and_commit_transaction():
    await APerson(name="Roger").save()

    await adb.begin()
    await APerson(name="Terry S").save()
    await adb.rollback()

    assert len(await APerson.nodes) == 1

    await adb.begin()
    await APerson(name="Terry S").save()
    await adb.commit()

    assert len(await APerson.nodes) == 2


@adb.transaction
async def in_a_tx(*names):
    for n in names:
        await APerson(name=n).save()


@mark_async_test
async def test_transaction_decorator():
    await adb.install_labels(APerson)

    # should work
    await in_a_tx("Roger")

    # should bail but raise correct error
    with raises(UniqueProperty):
        await in_a_tx("Jim", "Roger")

    assert "Jim" not in [p.name for p in await APerson.nodes]


@mark_async_test
async def test_transaction_as_a_context():
    async with adb.transaction:
        await APerson(name="Tim").save()

    assert await APerson.nodes.filter(name="Tim")

    with raises(UniqueProperty):
        async with adb.transaction:
            await APerson(name="Tim").save()


@mark_async_test
async def test_query_inside_transaction():
    async with adb.transaction:
        await APerson(name="Alice").save()
        await APerson(name="Bob").save()

        assert len([p.name for p in await APerson.nodes]) == 2


@mark_async_test
async def test_read_transaction():
    await APerson(name="Johnny").save()

    async with adb.read_transaction:
        people = await APerson.nodes
        assert people

    with raises(TransactionError):
        async with adb.read_transaction:
            with raises(ClientError) as e:
                await APerson(name="Gina").save()
            assert e.value.code == "Neo.ClientError.Statement.AccessMode"


@mark_async_test
async def test_write_transaction():
    async with adb.write_transaction:
        await APerson(name="Amelia").save()

    amelia = await APerson.nodes.get(name="Amelia")
    assert amelia


@mark_async_test
async def double_transaction():
    await adb.begin()
    with raises(SystemError, match=r"Transaction in progress"):
        await adb.begin()

    await adb.rollback()


@adb.transaction.with_bookmark
async def in_a_tx_with_bookmark(*names):
    for n in names:
        await APerson(name=n).save()


@mark_async_test
async def test_bookmark_transaction_decorator():
    # should work
    result, bookmarks = await in_a_tx_with_bookmark("Ruth", bookmarks=None)
    assert result is None
    assert isinstance(bookmarks, Bookmarks)

    # should bail but raise correct error
    with raises(UniqueProperty):
        await in_a_tx_with_bookmark("Jane", "Ruth")

    assert "Jane" not in [p.name for p in await APerson.nodes]


@mark_async_test
async def test_bookmark_transaction_as_a_context():
    async with adb.transaction as transaction:
        await APerson(name="Tanya").save()
    assert isinstance(transaction.last_bookmark, Bookmarks)

    assert await APerson.nodes.filter(name="Tanya")

    with raises(UniqueProperty):
        async with adb.transaction as transaction:
            await APerson(name="Tanya").save()
    assert not hasattr(transaction, "last_bookmark")


@pytest.fixture
def spy_on_db_begin(monkeypatch):
    spy_calls = []
    original_begin = adb.begin

    def begin_spy(*args, **kwargs):
        spy_calls.append((args, kwargs))
        return original_begin(*args, **kwargs)

    monkeypatch.setattr(adb, "begin", begin_spy)
    return spy_calls


@mark_async_test
async def test_bookmark_passed_in_to_context(spy_on_db_begin):
    transaction = adb.transaction
    async with transaction:
        pass

    assert (spy_on_db_begin)[-1] == ((), {"access_mode": None, "bookmarks": None})
    last_bookmark = transaction.last_bookmark

    transaction.bookmarks = last_bookmark
    async with transaction:
        pass
    assert spy_on_db_begin[-1] == (
        (),
        {"access_mode": None, "bookmarks": last_bookmark},
    )


@mark_async_test
async def test_query_inside_bookmark_transaction():
    async with adb.transaction as transaction:
        await APerson(name="Alice").save()
        await APerson(name="Bob").save()

        assert len([p.name for p in await APerson.nodes]) == 2

    assert isinstance(transaction.last_bookmark, Bookmarks)



import os

from alabaster import _version as version


def get_path():
    """
    Shortcut for users whose theme is next to their conf.py.
    """
    # Theme directory is defined as our parent directory
    return os.path.abspath(os.path.dirname(os.path.dirname(__file__)))


def update_context(app, pagename, templatename, context, doctree):
    context["alabaster_version"] = version.__version__


def setup(app):
    app.connect("html-page-context", update_context)



"""
.. _neomodel_inspect_database:

``neomodel_inspect_database``
-----------------------------

::

    usage: neomodel_inspect_database [-h] [--db bolt://neo4j:neo4j@localhost:7687] [--write-to <someapp/models.py> ...]
    
    Connects to a Neo4j database and inspects existing nodes and relationships.
    Infers the schema of the database and generates Python class definitions.

    If a connection URL is not specified, the tool will look up the environment 
    variable NEO4J_BOLT_URL. If that environment variable is not set, the tool
    will attempt to connect to the default URL bolt://neo4j:neo4j@localhost:7687

    If a file is specified, the tool will write the class definitions to that file.
    If no file is specified, the tool will print the class definitions to stdout.

    Note : this script only has a synchronous mode.
    
    options:
        -h, --help            show this help message and exit
        --db bolt://neo4j:neo4j@localhost:7687
                            Neo4j Server URL
        -T, --write-to someapp/models.py
                            File where to write output.
        --no-rel-props        Do not inspect relationship properties
        --no-rel-cardinality  Do not infer relationship cardinality
"""

import argparse
import string
import textwrap
from os import environ
from typing import Any

from neomodel.sync_.core import db

IMPORTS = []


def parse_prop_class(prop_type):
    _import = ""
    prop_class = ""
    if prop_type.startswith("LIST OF"):
        _import = "ArrayProperty"
        prop_class = (
            f"ArrayProperty({parse_prop_class(prop_type.replace('LIST OF ', ''))})"
        )
    else:
        if prop_type == "STRING":
            _import = "StringProperty"
            prop_class = f"{_import}("
        elif prop_type == "BOOLEAN":
            _import = "BooleanProperty"
            prop_class = f"{_import}("
        elif prop_type == "DATE":
            _import = "DateProperty"
            prop_class = f"{_import}("
        elif prop_type == "DATE_TIME":
            _import = "DateTimeProperty"
            prop_class = f"{_import}("
        elif prop_type == "INTEGER":
            _import = "IntegerProperty"
            prop_class = f"{_import}("
        elif prop_type == "FLOAT":
            _import = "FloatProperty"
            prop_class = f"{_import}("
        elif prop_type == "POINT":
            _import = "PointProperty"
            prop_class = f"{_import}(crs='wgs-84'"

    if _import not in IMPORTS:
        IMPORTS.append(_import)
    return prop_class


class NodeInspector:
    @staticmethod
    def get_properties_for_label(label):
        query = f"""
          MATCH (n:`{label}`)
          WITH DISTINCT keys(n) as properties, head(collect(n)) AS sampleNode
          ORDER BY size(properties) DESC
          RETURN apoc.meta.cypher.types(properties(sampleNode)) AS properties LIMIT 1
        """
        result, _ = db.cypher_query(query)
        if result is not None and len(result) > 0:
            return result[0][0]

    @staticmethod
    def get_constraints_for_label(label):
        constraints, meta_constraints = db.cypher_query(
            f"SHOW CONSTRAINTS WHERE entityType='NODE' AND '{label}' IN labelsOrTypes AND type='UNIQUENESS'"
        )
        constraints_as_dict = [dict(zip(meta_constraints, row)) for row in constraints]
        constrained_properties = [
            item.get("properties")[0]
            for item in constraints_as_dict
            if len(item.get("properties")) == 1
        ]
        return constrained_properties

    @staticmethod
    def get_indexed_properties_for_label(label):
        if db.version_is_higher_than("5.0"):
            indexes, meta_indexes = db.cypher_query(
                f"SHOW INDEXES WHERE entityType='NODE' AND '{label}' IN labelsOrTypes AND type='RANGE' AND owningConstraint IS NULL"
            )
        else:
            indexes, meta_indexes = db.cypher_query(
                f"SHOW INDEXES WHERE entityType='NODE' AND '{label}' IN labelsOrTypes AND type='BTREE' AND uniqueness='NONUNIQUE'"
            )
        indexes_as_dict = [dict(zip(meta_indexes, row)) for row in indexes]
        indexed_properties = [
            item.get("properties")[0]
            for item in indexes_as_dict
            if len(item.get("properties")) == 1
        ]
        return indexed_properties


class RelationshipInspector:
    @classmethod
    def outgoing_relationships(cls, start_label, get_properties: bool = True):
        if get_properties:
            query = f"""
                MATCH (n:`{start_label}`)-[r]->(m)
                WITH DISTINCT type(r) as rel_type, head(labels(m)) AS target_label, keys(r) AS properties, head(collect(r)) AS sampleRel
                ORDER BY size(properties) DESC
                RETURN DISTINCT rel_type, target_label, collect(DISTINCT apoc.meta.cypher.types(properties(sampleRel)))[0] AS properties
            """
        else:
            query = f"""
                MATCH (n:`{start_label}`)-[r]->(m)
                WITH DISTINCT type(r) as rel_type, head(labels(m)) AS target_label
                RETURN rel_type, target_label, {{}} AS properties
            """
        result, _ = db.cypher_query(query)
        return [(record[0], record[1], record[2]) for record in result]

    @staticmethod
    def get_constraints_for_type(rel_type):
        constraints, meta_constraints = db.cypher_query(
            f"SHOW CONSTRAINTS WHERE entityType='RELATIONSHIP' AND '{rel_type}' IN labelsOrTypes AND type='RELATIONSHIP_UNIQUENESS'"
        )
        constraints_as_dict = [dict(zip(meta_constraints, row)) for row in constraints]
        constrained_properties = [
            item.get("properties")[0]
            for item in constraints_as_dict
            if len(item.get("properties")) == 1
        ]
        return constrained_properties

    @staticmethod
    def get_indexed_properties_for_type(rel_type):
        if db.version_is_higher_than("5.0"):
            indexes, meta_indexes = db.cypher_query(
                f"SHOW INDEXES WHERE entityType='RELATIONSHIP' AND '{rel_type}' IN labelsOrTypes AND type='RANGE' AND owningConstraint IS NULL"
            )
        else:
            indexes, meta_indexes = db.cypher_query(
                f"SHOW INDEXES WHERE entityType='RELATIONSHIP' AND '{rel_type}' IN labelsOrTypes AND type='BTREE' AND uniqueness='NONUNIQUE'"
            )
        indexes_as_dict = [dict(zip(meta_indexes, row)) for row in indexes]
        indexed_properties = [
            item.get("properties")[0]
            for item in indexes_as_dict
            if len(item.get("properties")) == 1
        ]
        return indexed_properties

    @staticmethod
    def infer_cardinality(rel_type, start_label):
        range_start_query = f"MATCH (n:`{start_label}`) WHERE NOT EXISTS ((n)-[:`{rel_type}`]->()) WITH n LIMIT 1 RETURN count(n)"
        result, _ = db.cypher_query(range_start_query)
        is_start_zero = result[0][0] > 0

        range_end_query = f"""
            MATCH (n:`{start_label}`)-[rel:`{rel_type}`]->()
            WITH n, count(rel) AS rel_count
            WHERE rel_count > 1
            WITH n LIMIT 1
            RETURN count(n)
        """
        result, _ = db.cypher_query(range_end_query)
        is_end_one = result[0][0] == 0

        cardinality = "Zero" if is_start_zero else "One"
        cardinality += "OrOne" if is_end_one and is_start_zero else "OrMore"

        if cardinality not in IMPORTS:
            IMPORTS.append(cardinality)

        return cardinality


def get_node_labels():
    query = "CALL db.labels()"
    result, _ = db.cypher_query(query)
    return [record[0] for record in result]


def build_prop_string(unique_properties, indexed_properties, prop, prop_type):
    is_unique = prop in unique_properties
    is_indexed = prop in indexed_properties
    index_str = ""
    if is_unique:
        index_str = "unique_index=True"
    elif is_indexed:
        index_str = "index=True"
    return f"    {prop.replace(' ', '_')} = {parse_prop_class(prop_type)}{index_str})\n"


def clean_class_member_key(key):
    return key.replace(" ", "_")


def generate_rel_class_name(rel_type):
    # Relationship type best practices are like FRIENDS_WITH
    # We want to convert that to FriendsWithRel
    return string.capwords(rel_type.replace("_", " ")).replace(" ", "") + "Rel"


def parse_imports():
    imports = ""
    if IMPORTS:
        special_imports = ""
        if "PointProperty" in IMPORTS:
            IMPORTS.remove("PointProperty")
            special_imports += (
                "from neomodel.contrib.spatial_properties import PointProperty\n"
            )
        imports = f"from neomodel import {', '.join(IMPORTS)}\n" + special_imports
    return imports


def build_rel_type_definition(
    label, outgoing_relationships, defined_rel_types, infer_cardinality: bool = True
):
    class_definition_append = ""
    rel_type_definitions = ""

    for rel in outgoing_relationships:
        rel_type = rel[0]
        rel_name = rel_type.lower()
        target_label = rel[1]
        rel_props = rel[2]

        unique_properties = (
            RelationshipInspector.get_constraints_for_type(rel_type)
            if db.version_is_higher_than("5.7")
            else []
        )
        indexed_properties = RelationshipInspector.get_indexed_properties_for_type(
            rel_type
        )

        cardinality_string = ""
        if infer_cardinality:
            cardinality = RelationshipInspector.infer_cardinality(rel_type, label)
            cardinality_string += f", cardinality={cardinality}"

        class_definition_append += f'    {clean_class_member_key(rel_name)} = RelationshipTo("{target_label}", "{rel_type}"{cardinality_string}'

        if rel_props and rel_type not in defined_rel_types:
            rel_model_name = generate_rel_class_name(rel_type)
            class_definition_append += f', model="{rel_model_name}"'
            rel_type_definitions = f"\n\nclass {rel_model_name}(StructuredRel):\n"
            rel_type_definitions += "".join(
                [
                    build_prop_string(
                        unique_properties, indexed_properties, prop, prop_type
                    )
                    for prop, prop_type in rel_props.items()
                ]
            )

        class_definition_append += ")\n"

    class_definition_append += rel_type_definitions

    return class_definition_append


def inspect_database(
    bolt_url,
    get_relationship_properties: bool = True,
    infer_relationship_cardinality: bool = True,
):
    # Connect to the database
    print(f"Connecting to {bolt_url}")
    db.set_connection(bolt_url)

    node_labels: list[Any] = get_node_labels()
    defined_rel_types: list[str] = []
    class_definitions: str = ""

    if node_labels:
        IMPORTS.append("StructuredNode")

    for label in node_labels:
        class_name = clean_class_member_key(label)
        properties = NodeInspector.get_properties_for_label(label)
        unique_properties = NodeInspector.get_constraints_for_label(label)
        indexed_properties = NodeInspector.get_indexed_properties_for_label(label)

        class_definition = f"class {class_name}(StructuredNode):\n"
        if properties:
            class_definition += "".join(
                [
                    build_prop_string(
                        unique_properties, indexed_properties, prop, prop_type
                    )
                    for prop, prop_type in properties.items()
                ]
            )

        outgoing_relationships = RelationshipInspector.outgoing_relationships(
            label, get_relationship_properties
        )

        if outgoing_relationships and "StructuredRel" not in IMPORTS:
            IMPORTS.append("RelationshipTo")
            # No rel properties = no rel classes
            # Then StructuredRel import is not needed
            if get_relationship_properties:
                IMPORTS.append("StructuredRel")

        class_definition += build_rel_type_definition(
            label,
            outgoing_relationships,
            defined_rel_types,
            infer_relationship_cardinality,
        )

        if not properties and not outgoing_relationships:
            class_definition += "    pass\n"

        class_definition += "\n\n"

        class_definitions += class_definition

    # Finally, parse imports
    imports = parse_imports()
    output = "\n".join([imports, class_definitions])

    return output


def main():
    # Parse command line arguments
    parser = argparse.ArgumentParser(
        description=textwrap.dedent(
            """
            Connects to a Neo4j database and inspects existing nodes and relationships.
            Infers the schema of the database and generates Python class definitions.

            If a connection URL is not specified, the tool will look up the environment 
            variable NEO4J_BOLT_URL. If that environment variable is not set, the tool
            will attempt to connect to the default URL bolt://neo4j:neo4j@localhost:7687

            If a file is specified, the tool will write the class definitions to that file.
            If no file is specified, the tool will print the class definitions to stdout.
            """
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "--db",
        metavar="bolt://neo4j:neo4j@localhost:7687",
        dest="neo4j_bolt_url",
        type=str,
        default="",
        help="Neo4j Server URL",
    )

    parser.add_argument(
        "-T",
        "--write-to",
        metavar="<someapp/models.py>",
        type=str,
        help="File where to write output.",
    )

    parser.add_argument(
        "--no-rel-props",
        dest="get_relationship_properties",
        action="store_false",
        help="Do not inspect relationship properties",
    )

    parser.add_argument(
        "--no-rel-cardinality",
        dest="infer_relationship_cardinality",
        action="store_false",
        help="Do not infer relationship cardinality",
    )

    args = parser.parse_args()

    bolt_url = args.neo4j_bolt_url
    if len(bolt_url) == 0:
        bolt_url = environ.get("NEO4J_BOLT_URL", "bolt://neo4j:neo4j@localhost:7687")

    # If a file is specified, write to that file
    # First try to open the file for writing to make sure it is writable
    # Before connecting to the database
    if args.write_to:
        with open(args.write_to, "w") as file:
            output = inspect_database(
                bolt_url=bolt_url,
                get_relationship_properties=args.get_relationship_properties,
                infer_relationship_cardinality=args.infer_relationship_cardinality,
            )
            print(f"Writing to {args.write_to}")
            file.write(output)
    # If no file is specified, print to stdout
    else:
        print(
            inspect_database(
                bolt_url=bolt_url,
                get_relationship_properties=args.get_relationship_properties,
                infer_relationship_cardinality=args.infer_relationship_cardinality,
            )
        )


if __name__ == "__main__":
    main()



from typing import Any, Optional, Type, Union


class NeomodelException(Exception):
    """
    A base class that identifies all exceptions raised by :mod:`neomodel`.
    """

    pass


class AttemptedCardinalityViolation(NeomodelException):
    """
    Attempted to alter the database state against the cardinality definitions.

    Example: a relationship of type `One` trying to connect a second node.
    """

    pass


class CardinalityViolation(NeomodelException):
    """
    The state of database doesn't match the nodes cardinality definition.

    For example a relationship type `OneOrMore` returns no nodes.
    """

    def __init__(self, rel_manager: Any, actual: Union[int, str]):
        self.rel_manager = str(rel_manager)
        self.actual = str(actual)

    def __str__(self) -> str:
        return (
            f"CardinalityViolation: Expected: {self.rel_manager}, got: {self.actual}."
        )


class ModelDefinitionException(NeomodelException):
    """
    Abstract exception to handle error conditions related to the node-to-class registry.
    """

    def __init__(
        self,
        db_node_rel_class: Any,
        current_node_class_registry: dict[frozenset, Any],
        current_db_specific_node_class_registry: dict[str, dict],
    ):
        """
        Initialises the exception with the database node that caused the missmatch.

        :param db_node_rel_class: Depending on the concrete class, this is either a Neo4j driver node object
               from the DBMS, or a data model class from an application's hierarchy.
        :param current_node_class_registry: Dictionary that maps frozenset of
               node labels to model classes
        :param current_db_specific_node_class_registry: Dictionary that maps frozenset of
                node labels to model classes for specific databases
        """
        self.db_node_rel_class = db_node_rel_class
        self.current_node_class_registry = current_node_class_registry
        self.current_db_specific_node_class_registry = (
            current_db_specific_node_class_registry
        )

    def _get_node_class_registry_formatted(self) -> str:
        """
        Returns the current node class registry string formatted as a list of
        Labels --> <class to instantiate> entries.

        :return: str
        """
        output = "\n".join(
            map(
                lambda x: f"{','.join(x[0])} --> {x[1]}",
                self.current_node_class_registry.items(),
            )
        )
        for db, db_registry in self.current_db_specific_node_class_registry.items():
            output += f"\n\nDatabase-specific: {db}\n"
            output += "\n".join(
                list(
                    map(
                        lambda x: f"{','.join(x[0])} --> {x[1]}",
                        db_registry.items(),
                    )
                )
            )
        return output


class NodeClassNotDefined(ModelDefinitionException):
    """
    Raised when it is impossible to resolve a Neo4j driver Node to a
    data model object. This can happen as a result of a query returning
    nodes for which class definitions do exist but have not been imported
    or because the retrieved nodes contain more labels for a known class.

    In either of these cases the mismatch must be reported
    """

    def __str__(self) -> str:
        node_labels = ",".join(self.db_node_rel_class.labels)

        return f"Node with labels {node_labels} does not resolve to any of the known objects\n{self._get_node_class_registry_formatted()}\n"


class RelationshipClassNotDefined(ModelDefinitionException):
    """
    Raised when it is impossible to resolve a Neo4j driver Relationship to
    a data model object.
    """

    def __str__(self) -> str:
        relationship_type = self.db_node_rel_class.type
        return f"""
            Relationship of type {relationship_type} does not resolve to any of the known objects
            {self._get_node_class_registry_formatted()}
            Note that when using the fetch_relations method, the relationship type must be defined in the model, even if only defined to StructuredRel.
            Otherwise, neomodel will not be able to determine which relationship model to resolve into.
        """


class RelationshipClassRedefined(ModelDefinitionException):
    """
    Raised when an attempt is made to re-map a relationship label to a relationship model of an entirely different type
    """

    def __init__(
        self,
        db_rel_class_type: Any,
        current_node_class_registry: dict[frozenset, Any],
        current_db_specific_node_class_registry: dict[str, dict],
        remapping_to_class: Any,
    ):
        """
        Initialises a relationship redefinition exception with the required data as follows:

        :param db_rel_class_type: The type of the relationship that caused the error.
        :type db_rel_class_type: str (The label of the relationship that caused the error)
        :param current_node_class_registry: The current db object's node-class registry.
        :param current_db_specific_node_class_registry: The current db object's node-class registry for specific databases.
        :type current_node_class_registry: dict
        :param remapping_to_class: The relationship class the relationship type was attempted to be redefined to.
        :type remapping_to_class: class
        """
        super().__init__(
            db_rel_class_type,
            current_node_class_registry,
            current_db_specific_node_class_registry,
        )
        self.remapping_to_class = remapping_to_class

    def __str__(self) -> str:
        relationship_type = self.db_node_rel_class
        return f"Relationship of type {relationship_type} redefined as {self.remapping_to_class}.\n{self._get_node_class_registry_formatted()}\n"


class NodeClassAlreadyDefined(ModelDefinitionException):
    """
    Raised when an attempt is made to re-map a set of labels to a class
    that already has a mapping within the node-to-class registry.
    """

    def __str__(self) -> str:
        node_class_labels = ",".join(self.db_node_rel_class.inherited_labels())

        return f"Class {self.db_node_rel_class.__module__}.{self.db_node_rel_class.__name__} with labels {node_class_labels} already defined:\n{self._get_node_class_registry_formatted()}\n"


class ConstraintValidationFailed(ValueError, NeomodelException):
    def __init__(self, msg: str):
        self.message = msg


class DeflateError(ValueError, NeomodelException):
    def __init__(self, key: str, cls: Any, msg: str, obj: Any):
        self.property_name = key
        self.node_class = cls
        self.msg = msg
        self.obj = repr(obj)

    def __str__(self) -> str:
        return f"Attempting to deflate property '{self.property_name}' on {self.obj} of class '{self.node_class.__name__}': {self.msg}"


class DoesNotExist(NeomodelException):
    _model_class: Optional[Type] = None
    """
    This class property refers the model class that a subclass of this class
    belongs to. It is set by :class:`~neomodel.core.NodeMeta`.
    """

    def __init__(self, msg: str):
        if self._model_class is None:
            raise RuntimeError("This class hasn't been setup properly.")
        self.message: str = msg
        super().__init__(self, msg)

    def __reduce__(self) -> tuple:
        return _unpickle_does_not_exist, (self._model_class, self.message)


def _unpickle_does_not_exist(_model_class: Any, message: str) -> DoesNotExist:
    return _model_class.DoesNotExist(message)


class InflateConflict(NeomodelException):
    def __init__(self, cls: Any, key: str, value: Any, nid: str):
        self.cls_name = cls.__name__
        self.property_name = key
        self.value = value
        self.nid = nid

    def __str__(self) -> str:
        return f"Found conflict with node {self.nid}, has property '{self.property_name}' with value '{self.value}' although class {self.cls_name} already has a property '{self.property_name}'"


class InflateError(ValueError, NeomodelException):
    def __init__(self, key: str, cls: Any, msg: str, obj: Optional[Any] = None):
        self.property_name = key
        self.node_class = cls
        self.msg = msg
        self.obj = repr(obj)

    def __str__(self) -> str:
        return f"Attempting to inflate property '{self.property_name}' on {self.obj} of class '{self.node_class.__name__}': {self.msg}"


class DeflateConflict(InflateConflict):
    def __init__(self, cls: Any, key: str, value: Any, nid: str):
        self.cls_name = cls.__name__
        self.property_name = key
        self.value = value
        self.nid = nid if nid else "(unsaved)"

    def __str__(self) -> str:
        return f"Found trying to set property '{self.property_name}' with value '{self.value}' on node {self.nid} although class {self.cls_name} already has a property '{self.property_name}'"


class MultipleNodesReturned(ValueError, NeomodelException):
    def __init__(self, msg: str):
        self.message = msg


class NotConnected(NeomodelException):
    def __init__(self, action: str, node1: Any, node2: Any):
        self.action = action
        self.node1 = node1
        self.node2 = node2

    def __str__(self) -> str:
        return f"Error performing '{self.action}' - Node {self.node1.element_id} of type '{self.node1.__class__.__name__}' is not connected to {self.node2.element_id} of type '{self.node2.__class__.__name__}'."


class RequiredProperty(NeomodelException):
    def __init__(self, key: str, cls: Any):
        self.property_name = key
        self.node_class = cls

    def __str__(self) -> str:
        return f"property '{self.property_name}' on objects of class {self.node_class.__name__}"


class UniqueProperty(ConstraintValidationFailed):
    def __init__(self, msg: str):
        self.message = msg


class FeatureNotSupported(NeomodelException):
    def __init__(self, msg: str):
        self.message = msg


__all__ = (
    AttemptedCardinalityViolation.__name__,
    CardinalityViolation.__name__,
    ConstraintValidationFailed.__name__,
    DeflateConflict.__name__,
    DeflateError.__name__,
    DoesNotExist.__name__,
    InflateConflict.__name__,
    InflateError.__name__,
    MultipleNodesReturned.__name__,
    NeomodelException.__name__,
    NotConnected.__name__,
    RequiredProperty.__name__,
    UniqueProperty.__name__,
    NodeClassNotDefined.__name__,
    NodeClassAlreadyDefined.__name__,
    RelationshipClassNotDefined.__name__,
    RelationshipClassRedefined.__name__,
    FeatureNotSupported.__name__,
)



from test._async_compat import mark_sync_test

from neomodel import StringProperty, StructuredNode

HOOKS_CALLED = {}


class HookTest(StructuredNode):
    name = StringProperty()

    def post_create(self):
        HOOKS_CALLED["post_create"] = 1

    def pre_save(self):
        HOOKS_CALLED["pre_save"] = 1

    def post_save(self):
        HOOKS_CALLED["post_save"] = 1

    def pre_delete(self):
        HOOKS_CALLED["pre_delete"] = 1

    def post_delete(self):
        HOOKS_CALLED["post_delete"] = 1


@mark_sync_test
def test_hooks():
    ht = HookTest(name="k").save()
    ht.delete()
    assert "pre_save" in HOOKS_CALLED
    assert "post_save" in HOOKS_CALLED
    assert "post_create" in HOOKS_CALLED
    assert "pre_delete" in HOOKS_CALLED
    assert "post_delete" in HOOKS_CALLED



from collections.abc import Iterator

from neo4j.graph import Path

from neomodel.async_.core import AsyncStructuredNode, adb
from neomodel.async_.relationship import AsyncStructuredRel


class AsyncNeomodelPath(object):
    """
    Represents paths within neomodel.

    This object is instantiated when you include whole paths in your ``cypher_query()``
    result sets and turn ``resolve_objects`` to True.

    That is, any query of the form:
    ::

        MATCH p=(:SOME_NODE_LABELS)-[:SOME_REL_LABELS]-(:SOME_OTHER_NODE_LABELS) return p

    ``NeomodelPath`` are simple objects that reference their nodes and relationships, each of which is already
    resolved to their neomodel objects if such mapping is possible.


    :param nodes: Neomodel nodes appearing in the path in order of appearance.
    :param relationships: Neomodel relationships appearing in the path in order of appearance.
    :type nodes: list[StructuredNode]
    :type relationships: list[StructuredRel]
    """

    def __init__(self, a_neopath: Path):
        self._nodes: list[AsyncStructuredNode] = []
        self._relationships: list[AsyncStructuredRel] = []

        for a_node in a_neopath.nodes:
            self._nodes.append(adb._object_resolution(a_node))

        for a_relationship in a_neopath.relationships:
            # This check is required here because if the relationship does not bear data
            # then it does not have an entry in the registry. In that case, we instantiate
            # an "unspecified" StructuredRel.
            rel_type = frozenset([a_relationship.type])
            if rel_type in adb._NODE_CLASS_REGISTRY:
                new_rel = adb._object_resolution(a_relationship)
            else:
                new_rel = AsyncStructuredRel.inflate(a_relationship)
            self._relationships.append(new_rel)

    def __repr__(self) -> str:
        return "<Path start=%r end=%r size=%s>" % (
            self.start_node,
            self.end_node,
            len(self),
        )

    def __len__(self) -> int:
        return len(self._relationships)

    def __iter__(self) -> Iterator[AsyncStructuredRel]:
        return iter(self._relationships)

    @property
    def nodes(self) -> list[AsyncStructuredNode]:
        return self._nodes

    @property
    def start_node(self) -> AsyncStructuredNode:
        """The first :class:`.StructuredNode` in this path."""
        return self._nodes[0]

    @property
    def end_node(self) -> AsyncStructuredNode:
        """The last :class:`.StructuredNode` in this path."""
        return self._nodes[-1]

    @property
    def relationships(self) -> list[AsyncStructuredRel]:
        return self._relationships



"""Custom types used for annotations."""

from typing import Any, Optional, TypedDict

Transformation = TypedDict(
    "Transformation",
    {
        "source": Any,
        "source_prop": Optional[str],
        "include_in_return": Optional[bool],
    },
)


Subquery = TypedDict(
    "Subquery",
    {
        "query": str,
        "query_params": dict,
        "return_set": list[str],
        "initial_context": Optional[list[Any]],
    },
)



from test._async_compat import mark_sync_test

import pytest

from neomodel import db
from neomodel.util import version_tag_to_integer


@mark_sync_test
def test_version_awareness():
    db_version = db.database_version
    if db_version != "5.7.0":
        pytest.skip("Testing a specific database version")
    assert db_version == "5.7.0"
    assert db.version_is_higher_than("5.7")
    assert db.version_is_higher_than("5.6.0")
    assert db.version_is_higher_than("5")
    assert db.version_is_higher_than("4")

    assert not db.version_is_higher_than("5.8")


@mark_sync_test
def test_edition_awareness():
    db_edition = db.database_edition
    if db_edition == "enterprise":
        assert db.edition_is_enterprise()
    else:
        assert not db.edition_is_enterprise()


def test_version_tag_to_integer():
    assert version_tag_to_integer("5.7.1") == 50701
    assert version_tag_to_integer("5.1") == 50100
    assert version_tag_to_integer("5") == 50000
    assert version_tag_to_integer("5.14.1") == 51401
    assert version_tag_to_integer("5.14-aura") == 51400



from test._async_compat import mark_sync_test

import pytest

from neomodel import (
    DeflateConflict,
    InflateConflict,
    IntegerProperty,
    StringProperty,
    db,
)
from neomodel.contrib import SemiStructuredNode


class UserProf(SemiStructuredNode):
    email = StringProperty(unique_index=True, required=True)
    age = IntegerProperty(index=True)


class Dummy(SemiStructuredNode):
    pass


@mark_sync_test
def test_to_save_to_model_with_required_only():
    u = UserProf(email="dummy@test.com")
    assert u.save()


@mark_sync_test
def test_save_to_model_with_extras():
    u = UserProf(email="jim@test.com", age=3, bar=99)
    u.foo = True
    assert u.save()
    u = UserProf.nodes.get(age=3)
    assert u.foo is True
    assert u.bar == 99


@mark_sync_test
def test_save_empty_model():
    dummy = Dummy()
    assert dummy.save()


@mark_sync_test
def test_inflate_conflict():
    class PersonForInflateTest(SemiStructuredNode):
        name = StringProperty()
        age = IntegerProperty()

        def hello(self):
            print("Hi my names " + self.name)

    # An ok model
    props = {"name": "Jim", "age": 8, "weight": 11}
    db.cypher_query("CREATE (n:PersonForInflateTest $props)", {"props": props})
    jim = PersonForInflateTest.nodes.get(name="Jim")
    assert jim.name == "Jim"
    assert jim.age == 8
    assert jim.weight == 11

    # A model that conflicts on `hello`
    props = {"name": "Tim", "age": 8, "hello": "goodbye"}
    db.cypher_query("CREATE (n:PersonForInflateTest $props)", {"props": props})
    with pytest.raises(InflateConflict):
        PersonForInflateTest.nodes.get(name="Tim")


@mark_sync_test
def test_deflate_conflict():
    class PersonForDeflateTest(SemiStructuredNode):
        name = StringProperty()
        age = IntegerProperty()

        def hello(self):
            print("Hi my names " + self.name)

    tim = PersonForDeflateTest(name="Tim", age=8, weight=11).save()
    tim.hello = "Hi"
    with pytest.raises(DeflateConflict):
        tim.save()



from pygments.style import Style
from pygments.token import (
    Comment,
    Error,
    Generic,
    Keyword,
    Literal,
    Name,
    Number,
    Operator,
    Other,
    Punctuation,
    String,
    Whitespace,
)


# Originally based on FlaskyStyle which was based on 'tango'.
class Alabaster(Style):
    background_color = "#f8f8f8"  # doesn't seem to override CSS 'pre' styling?
    default_style = ""

    styles = {
        # No corresponding class for the following:
        # Text:                     "", # class:  ''
        Whitespace: "underline #f8f8f8",  # class: 'w'
        Error: "#a40000 border:#ef2929",  # class: 'err'
        Other: "#000000",  # class 'x'
        Comment: "italic #8f5902",  # class: 'c'
        Comment.Preproc: "noitalic",  # class: 'cp'
        Keyword: "bold #004461",  # class: 'k'
        Keyword.Constant: "bold #004461",  # class: 'kc'
        Keyword.Declaration: "bold #004461",  # class: 'kd'
        Keyword.Namespace: "bold #004461",  # class: 'kn'
        Keyword.Pseudo: "bold #004461",  # class: 'kp'
        Keyword.Reserved: "bold #004461",  # class: 'kr'
        Keyword.Type: "bold #004461",  # class: 'kt'
        Operator: "#582800",  # class: 'o'
        Operator.Word: "bold #004461",  # class: 'ow' - like keywords
        Punctuation: "bold #000000",  # class: 'p'
        # because special names such as Name.Class, Name.Function, etc.
        # are not recognized as such later in the parsing, we choose them
        # to look the same as ordinary variables.
        Name: "#000000",  # class: 'n'
        Name.Attribute: "#c4a000",  # class: 'na' - to be revised
        Name.Builtin: "#004461",  # class: 'nb'
        Name.Builtin.Pseudo: "#3465a4",  # class: 'bp'
        Name.Class: "#000000",  # class: 'nc' - to be revised
        Name.Constant: "#000000",  # class: 'no' - to be revised
        Name.Decorator: "#888",  # class: 'nd' - to be revised
        Name.Entity: "#ce5c00",  # class: 'ni'
        Name.Exception: "bold #cc0000",  # class: 'ne'
        Name.Function: "#000000",  # class: 'nf'
        Name.Property: "#000000",  # class: 'py'
        Name.Label: "#f57900",  # class: 'nl'
        Name.Namespace: "#000000",  # class: 'nn' - to be revised
        Name.Other: "#000000",  # class: 'nx'
        Name.Tag: "bold #004461",  # class: 'nt' - like a keyword
        Name.Variable: "#000000",  # class: 'nv' - to be revised
        Name.Variable.Class: "#000000",  # class: 'vc' - to be revised
        Name.Variable.Global: "#000000",  # class: 'vg' - to be revised
        Name.Variable.Instance: "#000000",  # class: 'vi' - to be revised
        Number: "#990000",  # class: 'm'
        Literal: "#000000",  # class: 'l'
        Literal.Date: "#000000",  # class: 'ld'
        String: "#4e9a06",  # class: 's'
        String.Backtick: "#4e9a06",  # class: 'sb'
        String.Char: "#4e9a06",  # class: 'sc'
        String.Doc: "italic #8f5902",  # class: 'sd' - like a comment
        String.Double: "#4e9a06",  # class: 's2'
        String.Escape: "#4e9a06",  # class: 'se'
        String.Heredoc: "#4e9a06",  # class: 'sh'
        String.Interpol: "#4e9a06",  # class: 'si'
        String.Other: "#4e9a06",  # class: 'sx'
        String.Regex: "#4e9a06",  # class: 'sr'
        String.Single: "#4e9a06",  # class: 's1'
        String.Symbol: "#4e9a06",  # class: 'ss'
        Generic: "#000000",  # class: 'g'
        Generic.Deleted: "#a40000",  # class: 'gd'
        Generic.Emph: "italic #000000",  # class: 'ge'
        Generic.Error: "#ef2929",  # class: 'gr'
        Generic.Heading: "bold #000080",  # class: 'gh'
        Generic.Inserted: "#00A000",  # class: 'gi'
        Generic.Output: "#888",  # class: 'go'
        Generic.Prompt: "#745334",  # class: 'gp'
        Generic.Strong: "bold #000000",  # class: 'gs'
        Generic.Subheading: "bold #800080",  # class: 'gu'
        Generic.Traceback: "bold #a40000",  # class: 'gt'
    }



"""
Provides a test case for issue 600 - "Pull request #592 cause an error in case of relationship inharitance".

The issue is outlined here: https://github.com/neo4j-contrib/neomodel/issues/600
"""

from test._async_compat import mark_sync_test

from neomodel import Relationship, StructuredNode, StructuredRel

try:
    basestring
except NameError:
    basestring = str


class Class1(StructuredRel):
    pass


class SubClass1(Class1):
    pass


class SubClass2(Class1):
    pass


class RelationshipDefinerSecondSibling(StructuredNode):
    rel_1 = Relationship(
        "RelationshipDefinerSecondSibling", "SOME_REL_LABEL", model=Class1
    )
    rel_2 = Relationship(
        "RelationshipDefinerSecondSibling", "SOME_REL_LABEL", model=SubClass1
    )
    rel_3 = Relationship(
        "RelationshipDefinerSecondSibling", "SOME_REL_LABEL", model=SubClass2
    )


class RelationshipDefinerParentLast(StructuredNode):
    rel_2 = Relationship(
        "RelationshipDefinerParentLast", "SOME_REL_LABEL", model=SubClass1
    )
    rel_3 = Relationship(
        "RelationshipDefinerParentLast", "SOME_REL_LABEL", model=SubClass2
    )
    rel_1 = Relationship(
        "RelationshipDefinerParentLast", "SOME_REL_LABEL", model=Class1
    )


# Test cases
@mark_sync_test
def test_relationship_definer_second_sibling():
    # Create a few entities
    A = (RelationshipDefinerSecondSibling.get_or_create({}))[0]
    B = (RelationshipDefinerSecondSibling.get_or_create({}))[0]
    C = (RelationshipDefinerSecondSibling.get_or_create({}))[0]

    # Add connections
    A.rel_1.connect(B)
    B.rel_2.connect(C)
    C.rel_3.connect(A)


@mark_sync_test
def test_relationship_definer_parent_last():
    # Create a few entities
    A = (RelationshipDefinerParentLast.get_or_create({}))[0]
    B = (RelationshipDefinerParentLast.get_or_create({}))[0]
    C = (RelationshipDefinerParentLast.get_or_create({}))[0]

    # Add connections
    A.rel_1.connect(B)
    B.rel_2.connect(C)
    C.rel_3.connect(A)



from test._async_compat import mark_sync_test

import pytest
from neo4j.exceptions import ClientError
from pytest import raises

from neomodel import db
from neomodel.exceptions import FeatureNotSupported


@mark_sync_test
def test_impersonate():
    if not db.edition_is_enterprise():
        pytest.skip("Skipping test for community edition")
    with db.impersonate(user="troygreene"):
        results, _ = db.cypher_query("RETURN 'Doo Wacko !'")
        assert results[0][0] == "Doo Wacko !"


@mark_sync_test
def test_impersonate_unauthorized():
    if not db.edition_is_enterprise():
        pytest.skip("Skipping test for community edition")
    with db.impersonate(user="unknownuser"):
        with raises(ClientError):
            _ = db.cypher_query("RETURN 'Gabagool'")


@mark_sync_test
def test_impersonate_multiple_transactions():
    if not db.edition_is_enterprise():
        pytest.skip("Skipping test for community edition")
    with db.impersonate(user="troygreene"):
        with db.transaction:
            results, _ = db.cypher_query("RETURN 'Doo Wacko !'")
            assert results[0][0] == "Doo Wacko !"

        with db.transaction:
            results, _ = db.cypher_query("SHOW CURRENT USER")
            assert results[0][0] == "troygreene"

    results, _ = db.cypher_query("SHOW CURRENT USER")
    assert results[0][0] == "neo4j"


@mark_sync_test
def test_impersonate_community():
    if db.edition_is_enterprise():
        pytest.skip("Skipping test for enterprise edition")
    with raises(FeatureNotSupported):
        with db.impersonate(user="troygreene"):
            _ = db.cypher_query("RETURN 'Gabagoogoo'")



import io
from test._async_compat import mark_sync_test
from unittest.mock import patch

from pytest import raises

from neomodel import (
    AttemptedCardinalityViolation,
    CardinalityViolation,
    IntegerProperty,
    One,
    OneOrMore,
    RelationshipFrom,
    RelationshipTo,
    StringProperty,
    StructuredNode,
    ZeroOrMore,
    ZeroOrOne,
    config,
    db,
)


class HairDryer(StructuredNode):
    version = IntegerProperty()


class ScrewDriver(StructuredNode):
    version = IntegerProperty()


class Car(StructuredNode):
    version = IntegerProperty()


class Monkey(StructuredNode):
    name = StringProperty()
    dryers = RelationshipTo("HairDryer", "OWNS_DRYER", cardinality=ZeroOrMore)
    driver = RelationshipTo("ScrewDriver", "HAS_SCREWDRIVER", cardinality=ZeroOrOne)
    car = RelationshipTo("Car", "HAS_CAR", cardinality=OneOrMore)
    toothbrush = RelationshipTo("ToothBrush", "HAS_TOOTHBRUSH", cardinality=One)


class ToothBrush(StructuredNode):
    name = StringProperty()


class Owner(StructuredNode):
    name = StringProperty(required=True)
    pets = RelationshipTo("Pet", "OWNS")


class Pet(StructuredNode):
    name = StringProperty(required=True)
    owner = RelationshipFrom("Owner", "OWNS", cardinality=One)


class Company(StructuredNode):
    name = StringProperty(required=True)
    employees = RelationshipTo("Employee", "EMPLOYS")


class Employee(StructuredNode):
    name = StringProperty(required=True)
    employer = RelationshipFrom("Company", "EMPLOYS", cardinality=ZeroOrOne)


class Manager(StructuredNode):
    name = StringProperty(required=True)
    assistant = RelationshipTo("Assistant", "MANAGES", cardinality=One)


class Assistant(StructuredNode):
    name = StringProperty(required=True)
    boss = RelationshipFrom("Manager", "MANAGES", cardinality=One)


@mark_sync_test
def test_cardinality_zero_or_more():
    m = Monkey(name="tim").save()
    assert m.dryers.all() == []
    single_dryer = m.driver.single()
    assert single_dryer is None
    h = HairDryer(version=1).save()

    m.dryers.connect(h)
    assert len(m.dryers.all()) == 1
    single_dryer = m.dryers.single()
    assert single_dryer.version == 1

    m.dryers.disconnect(h)
    assert m.dryers.all() == []
    single_dryer = m.driver.single()
    assert single_dryer is None

    h2 = HairDryer(version=2).save()
    m.dryers.connect(h)
    m.dryers.connect(h2)
    m.dryers.disconnect_all()
    assert m.dryers.all() == []
    single_dryer = m.driver.single()
    assert single_dryer is None


@mark_sync_test
def test_cardinality_zero_or_one():
    m = Monkey(name="bob").save()
    assert m.driver.all() == []
    single_driver = m.driver.single()
    assert m.driver.single() is None
    h = ScrewDriver(version=1).save()

    m.driver.connect(h)
    assert len(m.driver.all()) == 1
    single_driver = m.driver.single()
    assert single_driver.version == 1

    j = ScrewDriver(version=2).save()
    with raises(AttemptedCardinalityViolation):
        m.driver.connect(j)

    m.driver.reconnect(h, j)
    single_driver = m.driver.single()
    assert single_driver.version == 2

    # Forcing creation of a second ToothBrush to go around
    # AttemptedCardinalityViolation
    db.cypher_query(
        """
        MATCH (m:Monkey WHERE m.name="bob")
        CREATE (s:ScrewDriver {version:3})
        WITH m, s
        CREATE (m)-[:HAS_SCREWDRIVER]->(s)
    """
    )
    with raises(
        CardinalityViolation, match=r"CardinalityViolation: Expected: .*, got: 2."
    ):
        m.driver.all()


@mark_sync_test
def test_cardinality_one_or_more():
    m = Monkey(name="jerry").save()

    with raises(CardinalityViolation):
        m.car.all()

    with raises(CardinalityViolation):
        m.car.single()

    c = Car(version=2).save()
    m.car.connect(c)
    single_car = m.car.single()
    assert single_car.version == 2

    cars = m.car.all()
    assert len(cars) == 1

    with raises(AttemptedCardinalityViolation):
        m.car.disconnect(c)

    d = Car(version=3).save()
    m.car.connect(d)
    cars = m.car.all()
    assert len(cars) == 2

    m.car.disconnect(d)
    cars = m.car.all()
    assert len(cars) == 1


@mark_sync_test
def test_cardinality_one():
    m = Monkey(name="jerry").save()

    with raises(
        CardinalityViolation, match=r"CardinalityViolation: Expected: .*, got: none."
    ):
        m.toothbrush.all()

    with raises(CardinalityViolation):
        m.toothbrush.single()

    b = ToothBrush(name="Jim").save()
    m.toothbrush.connect(b)
    single_toothbrush = m.toothbrush.single()
    assert single_toothbrush.name == "Jim"

    x = ToothBrush(name="Jim").save()
    with raises(AttemptedCardinalityViolation):
        m.toothbrush.connect(x)

    with raises(AttemptedCardinalityViolation):
        m.toothbrush.disconnect(b)

    with raises(AttemptedCardinalityViolation):
        m.toothbrush.disconnect_all()

    # Forcing creation of a second ToothBrush to go around
    # AttemptedCardinalityViolation
    db.cypher_query(
        """
        MATCH (m:Monkey WHERE m.name="jerry")
        CREATE (t:ToothBrush {name:"Jim"})
        WITH m, t
        CREATE (m)-[:HAS_TOOTHBRUSH]->(t)
    """
    )
    with raises(
        CardinalityViolation, match=r"CardinalityViolation: Expected: .*, got: 2."
    ):
        m.toothbrush.all()

    jp = Monkey(name="Jean-Pierre")
    with raises(ValueError, match="Node has not been saved cannot connect!"):
        jp.toothbrush.connect(b)


@mark_sync_test
def test_relationship_from_one_cardinality_enforced():
    """
    Test that RelationshipFrom with cardinality=One prevents multiple connections.

    This addresses the GitHub issue where RelationshipFrom cardinality constraints
    were not being enforced.
    """
    # Setup
    config.SOFT_INVERSE_CARDINALITY_CHECK = False
    owner1 = Owner(name="Alice").save()
    owner2 = Owner(name="Bob").save()
    pet = Pet(name="Fluffy").save()

    # First connection should succeed
    owner1.pets.connect(pet)

    # Verify connection was established
    assert pet.owner.single() == owner1
    assert pet in owner1.pets.all()

    # Second connection should fail due to RelationshipFrom cardinality=One
    with raises(AttemptedCardinalityViolation):
        owner2.pets.connect(pet)

    stream = io.StringIO()
    with patch("sys.stdout", new=stream):
        config.SOFT_INVERSE_CARDINALITY_CHECK = True
        owner2.pets.connect(pet)
        assert pet in owner2.pets.all()

    console_output = stream.getvalue()
    assert "Cardinality violation detected" in console_output
    assert "Soft check is enabled so the relationship will be created" in console_output
    assert "strict check will be enabled by default in version 6.0" in console_output


@mark_sync_test
def test_relationship_from_zero_or_one_cardinality_enforced():
    """
    Test that RelationshipFrom with cardinality=ZeroOrOne prevents multiple connections.
    """
    # Setup
    config.SOFT_INVERSE_CARDINALITY_CHECK = False
    company1 = Company(name="TechCorp").save()
    company2 = Company(name="StartupInc").save()
    employee = Employee(name="John").save()

    # First connection should succeed
    company1.employees.connect(employee)

    # Verify connection was established
    assert employee.employer.single() == company1
    assert employee in company1.employees.all()

    # Second connection should fail due to RelationshipFrom cardinality=ZeroOrOne
    with raises(AttemptedCardinalityViolation):
        company2.employees.connect(employee)

    stream = io.StringIO()
    with patch("sys.stdout", new=stream):
        config.SOFT_INVERSE_CARDINALITY_CHECK = True
        company2.employees.connect(employee)
        assert employee in company2.employees.all()

    console_output = stream.getvalue()
    assert "Cardinality violation detected" in console_output
    assert "Soft check is enabled so the relationship will be created" in console_output
    assert "strict check will be enabled by default in version 6.0" in console_output


@mark_sync_test
def test_bidirectional_cardinality_validation():
    """
    Test that cardinality is validated on both ends when both sides have constraints.
    """
    # Setup
    config.SOFT_INVERSE_CARDINALITY_CHECK = False
    manager1 = Manager(name="Sarah").save()
    manager2 = Manager(name="David").save()
    assistant = Assistant(name="Alex").save()

    # First connection should succeed
    manager1.assistant.connect(assistant)

    # Verify bidirectional connection
    assert manager1.assistant.single() == assistant
    assert assistant.boss.single() == manager1

    # Second manager trying to connect to same assistant should fail
    with raises(AttemptedCardinalityViolation):
        manager2.assistant.connect(assistant)

    stream = io.StringIO()
    with patch("sys.stdout", new=stream):
        config.SOFT_INVERSE_CARDINALITY_CHECK = True
        manager2.assistant.connect(assistant)
        assert assistant in manager2.assistant.all()

    console_output = stream.getvalue()
    assert "Cardinality violation detected" in console_output
    assert "Soft check is enabled so the relationship will be created" in console_output
    assert "strict check will be enabled by default in version 6.0" in console_output



from test._async_compat import mark_async_test

import pytest

from neomodel import (
    DeflateConflict,
    InflateConflict,
    IntegerProperty,
    StringProperty,
    adb,
)
from neomodel.contrib import AsyncSemiStructuredNode


class UserProf(AsyncSemiStructuredNode):
    email = StringProperty(unique_index=True, required=True)
    age = IntegerProperty(index=True)


class Dummy(AsyncSemiStructuredNode):
    pass


@mark_async_test
async def test_to_save_to_model_with_required_only():
    u = UserProf(email="dummy@test.com")
    assert await u.save()


@mark_async_test
async def test_save_to_model_with_extras():
    u = UserProf(email="jim@test.com", age=3, bar=99)
    u.foo = True
    assert await u.save()
    u = await UserProf.nodes.get(age=3)
    assert u.foo is True
    assert u.bar == 99


@mark_async_test
async def test_save_empty_model():
    dummy = Dummy()
    assert await dummy.save()


@mark_async_test
async def test_inflate_conflict():
    class PersonForInflateTest(AsyncSemiStructuredNode):
        name = StringProperty()
        age = IntegerProperty()

        def hello(self):
            print("Hi my names " + self.name)

    # An ok model
    props = {"name": "Jim", "age": 8, "weight": 11}
    await adb.cypher_query("CREATE (n:PersonForInflateTest $props)", {"props": props})
    jim = await PersonForInflateTest.nodes.get(name="Jim")
    assert jim.name == "Jim"
    assert jim.age == 8
    assert jim.weight == 11

    # A model that conflicts on `hello`
    props = {"name": "Tim", "age": 8, "hello": "goodbye"}
    await adb.cypher_query("CREATE (n:PersonForInflateTest $props)", {"props": props})
    with pytest.raises(InflateConflict):
        await PersonForInflateTest.nodes.get(name="Tim")


@mark_async_test
async def test_deflate_conflict():
    class PersonForDeflateTest(AsyncSemiStructuredNode):
        name = StringProperty()
        age = IntegerProperty()

        def hello(self):
            print("Hi my names " + self.name)

    tim = await PersonForDeflateTest(name="Tim", age=8, weight=11).save()
    tim.hello = "Hi"
    with pytest.raises(DeflateConflict):
        await tim.save()



from test._async_compat import mark_sync_test

from neomodel import RelationshipTo, StructuredNode


class SomeModel(StructuredNode):
    test = RelationshipTo("SomeModel", "SELF")


@mark_sync_test
def test_len_relationship():
    t1 = SomeModel().save()
    t2 = SomeModel().save()

    t1.test.connect(t2)
    l = len(t1.test.all())

    assert l
    assert l == 1



import pickle
from test._async_compat import mark_sync_test

from neomodel import DoesNotExist, StringProperty, StructuredNode


class EPerson(StructuredNode):
    name = StringProperty(unique_index=True)


@mark_sync_test
def test_object_does_not_exist():
    try:
        EPerson.nodes.get(name="johnny")
    except EPerson.DoesNotExist as e:
        pickle_instance = pickle.dumps(e)
        assert pickle_instance
        assert pickle.loads(pickle_instance)
        assert isinstance(pickle.loads(pickle_instance), DoesNotExist)
    else:
        assert False, "Person.DoesNotExist not raised."


def test_pickle_does_not_exist():
    try:
        raise EPerson.DoesNotExist("My Test Message")
    except EPerson.DoesNotExist as e:
        pickle_instance = pickle.dumps(e)
        assert pickle_instance
        assert pickle.loads(pickle_instance)
        assert isinstance(pickle.loads(pickle_instance), DoesNotExist)



import functools
import inspect
import sys
from importlib import import_module
from typing import TYPE_CHECKING, Any, AsyncIterator, Callable, Optional, Union

from neomodel import config
from neomodel.async_.core import adb
from neomodel.async_.match import (
    AsyncNodeSet,
    AsyncTraversal,
    _rel_helper,
    _rel_merge_helper,
)
from neomodel.async_.relationship import AsyncStructuredRel
from neomodel.exceptions import NotConnected, RelationshipClassRedefined
from neomodel.util import (
    EITHER,
    INCOMING,
    OUTGOING,
    enumerate_traceback,
    get_graph_entity_properties,
)

if TYPE_CHECKING:
    from neomodel import AsyncStructuredNode
    from neomodel.async_.match import AsyncBaseSet


# check source node is saved and not deleted
def check_source(fn: Callable) -> Callable:
    fn_name = fn.func_name if hasattr(fn, "func_name") else fn.__name__

    @functools.wraps(fn)
    def checker(self: Any, *args: Any, **kwargs: Any) -> Callable:
        self.source._pre_action_check(self.name + "." + fn_name)
        return fn(self, *args, **kwargs)

    return checker


# checks if obj is a direct subclass, 1 level
def is_direct_subclass(obj: Any, classinfo: Any) -> bool:
    for base in obj.__bases__:
        if base == classinfo:
            return True
    return False


class AsyncRelationshipManager(object):
    """
    Base class for all relationships managed through neomodel.

    I.e the 'friends' object in  `user.friends.all()`
    """

    source: Any
    source_class: Any
    name: str
    definition: dict
    description: str = "relationship"

    def __init__(self, source: Any, key: str, definition: dict):
        self.source = source
        self.source_class = source.__class__
        self.name = key
        self.definition = definition

    def __str__(self) -> str:
        direction = "either"
        if self.definition["direction"] == OUTGOING:
            direction = "a outgoing"
        elif self.definition["direction"] == INCOMING:
            direction = "a incoming"

        return f"{self.description} in {direction} direction of type {self.definition['relation_type']} on node ({self.source.element_id}) of class '{self.source_class.__name__}'"

    def __await__(self) -> Any:
        return self.all().__await__()  # type: ignore[attr-defined]

    async def _check_cardinality(
        self, node: "AsyncStructuredNode", soft_check: bool = False
    ) -> None:
        """
        Check whether a new connection to a node would violate the cardinality
        of the relationship.

        :param node: The node that is being connected.
        :type: AsyncStructuredNode
        :raises: AttemptedCardinalityViolation
        """

    def _check_node(self, obj: type["AsyncStructuredNode"]) -> None:
        """check for valid node i.e correct class and is saved"""
        if not issubclass(type(obj), self.definition["node_class"]):
            raise ValueError(
                "Expected node of class " + self.definition["node_class"].__name__
            )
        if not hasattr(obj, "element_id"):
            raise ValueError("Can't perform operation on unsaved node " + repr(obj))

    @check_source
    async def connect(
        self, node: "AsyncStructuredNode", properties: Optional[dict[str, Any]] = None
    ) -> Optional[AsyncStructuredRel]:
        """
        Connect a node

        :param node:
        :param properties: for the new relationship
        :type: dict
        :return:
        """
        self._check_node(node)
        await self._check_cardinality(node)

        # Check for cardinality on the remote end.
        for rel_name, rel_def in node.defined_properties(
            rels=True, aliases=False, properties=False
        ).items():
            # In order to find the inverse relationship, we need to check
            # that the relationship type is the same, the direction is
            # opposite, and the node class is the same as the source.
            if (
                rel_def.definition["relation_type"] == self.definition["relation_type"]
                and rel_def.definition["direction"] != self.definition["direction"]
                and rel_def.definition["node_class"] == self.source_class
            ):
                # If we have found the inverse relationship, we need to check
                # its cardinality.
                inverse_rel = getattr(node, rel_name)
                await inverse_rel._check_cardinality(
                    self.source, soft_check=config.SOFT_INVERSE_CARDINALITY_CHECK
                )
                break

        if not self.definition["model"] and properties:
            raise NotImplementedError(
                "Relationship properties without using a relationship model "
                "is no longer supported."
            )

        params = {}
        rel_model = self.definition["model"]
        rel_prop = None

        if rel_model:
            rel_prop = {}
            # need to generate defaults etc to create fake instance
            tmp = rel_model(**properties) if properties else rel_model()
            # build params and place holders to pass to rel_helper
            for prop, val in rel_model.deflate(tmp.__properties__).items():
                if val is not None:
                    rel_prop[prop] = "$" + prop
                else:
                    rel_prop[prop] = None
                params[prop] = val

            if hasattr(tmp, "pre_save"):
                tmp.pre_save()

        new_rel = _rel_merge_helper(
            lhs="us",
            rhs="them",
            ident="r",
            relation_properties=rel_prop,
            **self.definition,
        )
        q = (
            f"MATCH (them), (us) WHERE {await adb.get_id_method()}(them)=$them and {await adb.get_id_method()}(us)=$self "
            "MERGE" + new_rel
        )

        params["them"] = await adb.parse_element_id(node.element_id)

        if not rel_model:
            await self.source.cypher(q, params)
            return None

        results = await self.source.cypher(q + " RETURN r", params)
        rel_ = results[0][0][0]
        rel_instance = self._set_start_end_cls(rel_model.inflate(rel_), node)

        if hasattr(rel_instance, "post_save"):
            rel_instance.post_save()

        return rel_instance

    @check_source
    async def replace(
        self, node: "AsyncStructuredNode", properties: Optional[dict[str, Any]] = None
    ) -> None:
        """
        Disconnect all existing nodes and connect the supplied node

        :param node:
        :param properties: for the new relationship
        :type: dict
        :return:
        """
        await self.disconnect_all()
        await self.connect(node, properties)

    @check_source
    async def relationship(
        self, node: "AsyncStructuredNode"
    ) -> Optional[AsyncStructuredRel]:
        """
        Retrieve the relationship object for this first relationship between self and node.

        :param node:
        :return: StructuredRel
        """
        self._check_node(node)
        my_rel = _rel_helper(lhs="us", rhs="them", ident="r", **self.definition)
        q = (
            "MATCH "
            + my_rel
            + f" WHERE {await adb.get_id_method()}(them)=$them and {await adb.get_id_method()}(us)=$self RETURN r LIMIT 1"
        )
        results = await self.source.cypher(
            q, {"them": await adb.parse_element_id(node.element_id)}
        )
        rels = results[0]
        if not rels:
            return None

        rel_model = self.definition.get("model") or AsyncStructuredRel

        return self._set_start_end_cls(rel_model.inflate(rels[0][0]), node)

    @check_source
    async def all_relationships(
        self, node: "AsyncStructuredNode"
    ) -> list[AsyncStructuredRel]:
        """
        Retrieve all relationship objects between self and node.

        :param node:
        :return: [StructuredRel]
        """
        self._check_node(node)

        my_rel = _rel_helper(lhs="us", rhs="them", ident="r", **self.definition)
        q = f"MATCH {my_rel} WHERE {await adb.get_id_method()}(them)=$them and {await adb.get_id_method()}(us)=$self RETURN r "
        results = await self.source.cypher(
            q, {"them": await adb.parse_element_id(node.element_id)}
        )
        rels = results[0]
        if not rels:
            return []

        rel_model = self.definition.get("model") or AsyncStructuredRel
        return [
            self._set_start_end_cls(rel_model.inflate(rel[0]), node) for rel in rels
        ]

    def _set_start_end_cls(
        self, rel_instance: AsyncStructuredRel, obj: "AsyncStructuredNode"
    ) -> AsyncStructuredRel:
        if self.definition["direction"] == INCOMING:
            rel_instance._start_node_class = obj.__class__
            rel_instance._end_node_class = self.source_class
        else:
            rel_instance._start_node_class = self.source_class
            rel_instance._end_node_class = obj.__class__
        return rel_instance

    @check_source
    async def reconnect(
        self, old_node: "AsyncStructuredNode", new_node: "AsyncStructuredNode"
    ) -> None:
        """
        Disconnect old_node and connect new_node copying over any properties on the original relationship.

        Useful for preventing cardinality violations

        :param old_node:
        :param new_node:
        :return: None
        """

        self._check_node(old_node)
        self._check_node(new_node)
        if old_node.element_id == new_node.element_id:
            return
        old_rel = _rel_helper(lhs="us", rhs="old", ident="r", **self.definition)

        # get list of properties on the existing rel
        old_node_element_id = await adb.parse_element_id(old_node.element_id)
        new_node_element_id = await adb.parse_element_id(new_node.element_id)
        result, _ = await self.source.cypher(
            f"""
                MATCH (us), (old) WHERE {await adb.get_id_method()}(us)=$self and {await adb.get_id_method()}(old)=$old
                MATCH {old_rel} RETURN r
            """,
            {"old": old_node_element_id},
        )
        if result:
            node_properties = get_graph_entity_properties(result[0][0])
            existing_properties = node_properties.keys()
        else:
            raise NotConnected("reconnect", self.source, old_node)

        # remove old relationship and create new one
        new_rel = _rel_merge_helper(lhs="us", rhs="new", ident="r2", **self.definition)
        q = (
            "MATCH (us), (old), (new) "
            f"WHERE {await adb.get_id_method()}(us)=$self and {await adb.get_id_method()}(old)=$old and {await adb.get_id_method()}(new)=$new "
            "MATCH " + old_rel
        )
        q += " MERGE" + new_rel

        # copy over properties if we have
        q += "".join([f" SET r2.{prop} = r.{prop}" for prop in existing_properties])
        q += " WITH r DELETE r"

        await self.source.cypher(
            q, {"old": old_node_element_id, "new": new_node_element_id}
        )

    @check_source
    async def disconnect(self, node: "AsyncStructuredNode") -> None:
        """
        Disconnect a node

        :param node:
        :return:
        """
        rel = _rel_helper(lhs="a", rhs="b", ident="r", **self.definition)
        q = f"""
                MATCH (a), (b) WHERE {await adb.get_id_method()}(a)=$self and {await adb.get_id_method()}(b)=$them
                MATCH {rel} DELETE r
            """
        await self.source.cypher(
            q, {"them": await adb.parse_element_id(node.element_id)}
        )

    @check_source
    async def disconnect_all(self) -> None:
        """
        Disconnect all nodes

        :return:
        """
        rhs = "b:" + self.definition["node_class"].__label__
        rel = _rel_helper(lhs="a", rhs=rhs, ident="r", **self.definition)
        q = (
            f"MATCH (a) WHERE {await adb.get_id_method()}(a)=$self MATCH "
            + rel
            + " DELETE r"
        )
        await self.source.cypher(q)

    @check_source
    def _new_traversal(self) -> AsyncTraversal:
        return AsyncTraversal(self.source, self.name, self.definition)

    # The methods below simply proxy the match engine.
    async def get(self, **kwargs: Any) -> AsyncNodeSet:
        """
        Retrieve a related node with the matching node properties.

        :param kwargs: same syntax as `NodeSet.filter()`
        :return: node
        """
        return await AsyncNodeSet(self._new_traversal()).get(**kwargs)

    async def get_or_none(self, **kwargs: dict) -> AsyncNodeSet:
        """
        Retrieve a related node with the matching node properties or return None.

        :param kwargs: same syntax as `NodeSet.filter()`
        :return: node
        """
        return await AsyncNodeSet(self._new_traversal()).get_or_none(**kwargs)

    def filter(self, *args: Any, **kwargs: dict) -> "AsyncBaseSet":
        """
        Retrieve related nodes matching the provided properties.

        :param args: a Q object
        :param kwargs: same syntax as `NodeSet.filter()`
        :return: NodeSet
        """
        return AsyncNodeSet(self._new_traversal()).filter(*args, **kwargs)

    def order_by(self, *props: Any) -> "AsyncBaseSet":
        """
        Order related nodes by specified properties

        :param props:
        :return: NodeSet
        """
        return AsyncNodeSet(self._new_traversal()).order_by(*props)

    def exclude(self, *args: Any, **kwargs: dict) -> "AsyncBaseSet":
        """
        Exclude nodes that match the provided properties.

        :param args: a Q object
        :param kwargs: same syntax as `NodeSet.filter()`
        :return: NodeSet
        """
        return AsyncNodeSet(self._new_traversal()).exclude(*args, **kwargs)

    async def is_connected(self, node: "AsyncStructuredNode") -> bool:
        """
        Check if a node is connected with this relationship type
        :param node:
        :return: bool
        """
        return await self._new_traversal().check_contains(node)

    async def single(self) -> Optional["AsyncStructuredNode"]:
        """
        Get a single related node or none.

        :return: StructuredNode
        """
        try:
            rels = await self
            return rels[0]
        except IndexError:
            return None

    def match(self, **kwargs: dict) -> AsyncNodeSet:
        """
        Return set of nodes who's relationship properties match supplied args

        :param kwargs: same syntax as `NodeSet.filter()`
        :return: NodeSet
        """
        return self._new_traversal().match(**kwargs)

    async def all(self) -> list:
        """
        Return all related nodes.

        :return: list
        """
        return await self._new_traversal().all()

    async def __aiter__(self) -> AsyncIterator:
        return self._new_traversal().__aiter__()

    async def get_len(self) -> int:
        return await self._new_traversal().get_len()

    async def check_bool(self) -> bool:
        return await self._new_traversal().check_bool()

    async def check_nonzero(self) -> bool:
        return self._new_traversal().check_nonzero()

    async def check_contains(self, obj: Any) -> bool:
        return self._new_traversal().check_contains(obj)

    async def get_item(self, key: Union[int, slice]) -> Any:
        return self._new_traversal().get_item(key)


class AsyncRelationshipDefinition:
    def __init__(
        self,
        relation_type: str,
        cls_name: str,
        direction: int,
        manager: type[AsyncRelationshipManager] = AsyncRelationshipManager,
        model: Optional[type[AsyncStructuredRel]] = None,
    ) -> None:
        self._validate_class(cls_name, model)

        current_frame = inspect.currentframe()

        frame_number = 3
        for i, frame in enumerate_traceback(current_frame):
            if cls_name in frame.f_globals:
                frame_number = i
                break
        self.module_name = sys._getframe(frame_number).f_globals["__name__"]
        if "__file__" in sys._getframe(frame_number).f_globals:
            self.module_file = sys._getframe(frame_number).f_globals["__file__"]
        self._raw_class = cls_name
        self.manager = manager
        self.definition = {
            "relation_type": relation_type,
            "direction": direction,
            "model": model,
        }

        if model is not None:
            # Relationships are easier to instantiate because
            # they cannot have multiple labels.
            # So, a relationship's type determines the class that should be
            # instantiated uniquely.
            # Here however, we still use a `frozenset([relation_type])`
            # to preserve the mapping type.
            label_set = frozenset([relation_type])
            try:
                # If the relationship mapping exists then it is attempted
                # to be redefined so that it applies to the same label.
                # In this case, it has to be ensured that the class
                # that is overriding the relationship is a descendant
                # of the already existing class.
                model_from_registry = adb._NODE_CLASS_REGISTRY[label_set]
                if not issubclass(model, model_from_registry):
                    is_parent = issubclass(model_from_registry, model)
                    if is_direct_subclass(model, AsyncStructuredRel) and not is_parent:
                        raise RelationshipClassRedefined(
                            relation_type,
                            adb._NODE_CLASS_REGISTRY,
                            adb._DB_SPECIFIC_CLASS_REGISTRY,
                            model,
                        )
                else:
                    adb._NODE_CLASS_REGISTRY[label_set] = model
            except KeyError:
                # If the mapping does not exist then it is simply created.
                adb._NODE_CLASS_REGISTRY[label_set] = model

    def _validate_class(
        self, cls_name: str, model: Optional[type[AsyncStructuredRel]] = None
    ) -> None:
        if not isinstance(cls_name, (str, object)):
            raise ValueError("Expected class name or class got " + repr(cls_name))

        if model and not issubclass(model, (AsyncStructuredRel,)):
            raise ValueError("model must be a StructuredRel")

    def lookup_node_class(self) -> None:
        if not isinstance(self._raw_class, str):
            self.definition["node_class"] = self._raw_class
        else:
            name = self._raw_class
            if name.find(".") == -1:
                module = self.module_name
            else:
                module, _, name = name.rpartition(".")

            if module not in sys.modules:
                # yet another hack to get around python semantics
                # __name__ is the namespace of the parent module for __init__.py files,
                # and the namespace of the current module for other .py files,
                # therefore there's a need to define the namespace differently for
                # these two cases in order for . in relative imports to work correctly
                # (i.e. to mean the same thing for both cases).
                # For example in the comments below, namespace == myapp, always
                if not hasattr(self, "module_file"):
                    raise ImportError(f"Couldn't lookup '{name}'")

                if "__init__.py" in self.module_file:
                    # e.g. myapp/__init__.py -[__name__]-> myapp
                    namespace = self.module_name
                else:
                    # e.g. myapp/models.py -[__name__]-> myapp.models
                    namespace = self.module_name.rpartition(".")[0]

                # load a module from a namespace (e.g. models from myapp)
                if module:
                    module = import_module(module, namespace).__name__
                # load the namespace itself (e.g. myapp)
                # (otherwise it would look like import . from myapp)
                else:
                    module = import_module(namespace).__name__
            self.definition["node_class"] = getattr(sys.modules[module], name)

    def build_manager(
        self, source: "AsyncStructuredNode", name: str
    ) -> AsyncRelationshipManager:
        self.lookup_node_class()
        return self.manager(source, name, self.definition)


class AsyncZeroOrMore(AsyncRelationshipManager):
    """
    A relationship of zero or more nodes (the default)
    """

    description = "zero or more relationships"


class AsyncRelationshipTo(AsyncRelationshipDefinition):
    def __init__(
        self,
        cls_name: str,
        relation_type: str,
        cardinality: type[AsyncRelationshipManager] = AsyncZeroOrMore,
        model: Optional[type[AsyncStructuredRel]] = None,
    ) -> None:
        super().__init__(
            relation_type, cls_name, OUTGOING, manager=cardinality, model=model
        )


class AsyncRelationshipFrom(AsyncRelationshipDefinition):
    def __init__(
        self,
        cls_name: str,
        relation_type: str,
        cardinality: type[AsyncRelationshipManager] = AsyncZeroOrMore,
        model: Optional[type[AsyncStructuredRel]] = None,
    ) -> None:
        super().__init__(
            relation_type, cls_name, INCOMING, manager=cardinality, model=model
        )


class AsyncRelationship(AsyncRelationshipDefinition):
    def __init__(
        self,
        cls_name: str,
        relation_type: str,
        cardinality: type[AsyncRelationshipManager] = AsyncZeroOrMore,
        model: Optional[type[AsyncStructuredRel]] = None,
    ) -> None:
        super().__init__(
            relation_type, cls_name, EITHER, manager=cardinality, model=model
        )



"""
Provides a test case for data types required by issue 374 - "Support for Point property type".

At the moment, only one new datatype is offered: NeomodelPoint

For more information please see: https://github.com/neo4j-contrib/neomodel/issues/374
"""

import os

import pytest
import shapely

import neomodel
import neomodel.contrib.spatial_properties
from neomodel.util import version_tag_to_integer


def check_and_skip_neo4j_least_version(required_least_neo4j_version, message):
    """
    Checks if the NEO4J_VERSION is at least `required_least_neo4j_version` and skips a test if not.

    WARNING: If the NEO4J_VERSION variable is not set, this function returns True, allowing the test to go ahead.

    :param required_least_neo4j_version: The least version to check. This must be the numberic representation of the
    version. That is: '3.4.0' would be passed as 340.
    :type required_least_neo4j_version: int
    :param message: An informative message as to why the calling test had to be skipped.
    :type message: str
    :return: A boolean value of True if the version reported is at least `required_least_neo4j_version`
    """
    if "NEO4J_VERSION" in os.environ:
        if (
            version_tag_to_integer(os.environ["NEO4J_VERSION"])
            < required_least_neo4j_version
        ):
            pytest.skip(
                "Neo4j version: {}. {}."
                "Skipping test.".format(os.environ["NEO4J_VERSION"], message)
            )


def basic_type_assertions(
    ground_truth, tested_object, test_description, check_neo4j_points=False
):
    """
    Tests that `tested_object` has been created as intended.

    :param ground_truth: The object as it is supposed to have been created.
    :type ground_truth: NeomodelPoint or neo4j.v1.spatial.Point
    :param tested_object: The object as it results from one of the contructors.
    :type tested_object: NeomodelPoint or neo4j.v1.spatial.Point
    :param test_description: A brief description of the test being performed.
    :type test_description: str
    :param check_neo4j_points: Whether to assert between NeomodelPoint or neo4j.v1.spatial.Point objects.
    :type check_neo4j_points: bool
    :return:
    """
    if check_neo4j_points:
        assert isinstance(
            tested_object, type(ground_truth)
        ), "{} did not return Neo4j Point".format(test_description)
        assert (
            tested_object.srid == ground_truth.srid
        ), "{} does not have the expected SRID({})".format(
            test_description, ground_truth.srid
        )
        assert len(tested_object) == len(
            ground_truth
        ), "Dimensionality mismatch. Expected {}, had {}".format(
            len(ground_truth.coords), len(tested_object.coords)
        )
    else:
        assert isinstance(
            tested_object, type(ground_truth)
        ), "{} did not return NeomodelPoint".format(test_description)
        assert (
            tested_object.crs == ground_truth.crs
        ), "{} does not have the expected CRS({})".format(
            test_description, ground_truth.crs
        )
        assert len(tested_object.coords[0]) == len(
            ground_truth.coords[0]
        ), "Dimensionality mismatch. Expected {}, had {}".format(
            len(ground_truth.coords[0]), len(tested_object.coords[0])
        )


# Object Construction
def test_coord_constructor():
    """
    Tests all the possible ways by which a NeomodelPoint can be instantiated successfully via passing coordinates.
    :return:
    """

    # Neo4j versions lower than 3.4.0 do not support Point. In that case, skip the test.
    check_and_skip_neo4j_least_version(
        340, "This version does not support spatial data types."
    )

    # Implicit cartesian point with coords
    ground_truth_object = neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0))
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0))
    basic_type_assertions(
        ground_truth_object,
        new_point,
        "Implicit 2d cartesian point instantiation",
    )

    ground_truth_object = neomodel.contrib.spatial_properties.NeomodelPoint(
        (0.0, 0.0, 0.0)
    )
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0, 0.0))
    basic_type_assertions(
        ground_truth_object,
        new_point,
        "Implicit 3d cartesian point instantiation",
    )

    # Explicit geographical point with coords
    ground_truth_object = neomodel.contrib.spatial_properties.NeomodelPoint(
        (0.0, 0.0), crs="wgs-84"
    )
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(
        (0.0, 0.0), crs="wgs-84"
    )
    basic_type_assertions(
        ground_truth_object,
        new_point,
        "Explicit 2d geographical point with tuple of coords instantiation",
    )

    ground_truth_object = neomodel.contrib.spatial_properties.NeomodelPoint(
        (0.0, 0.0, 0.0), crs="wgs-84-3d"
    )
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(
        (0.0, 0.0, 0.0), crs="wgs-84-3d"
    )
    basic_type_assertions(
        ground_truth_object,
        new_point,
        "Explicit 3d geographical point with tuple of coords instantiation",
    )

    # Cartesian point with named arguments
    ground_truth_object = neomodel.contrib.spatial_properties.NeomodelPoint(
        x=0.0, y=0.0
    )
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(x=0.0, y=0.0)
    basic_type_assertions(
        ground_truth_object,
        new_point,
        "Cartesian 2d point with named arguments",
    )

    ground_truth_object = neomodel.contrib.spatial_properties.NeomodelPoint(
        x=0.0, y=0.0, z=0.0
    )
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(x=0.0, y=0.0, z=0.0)
    basic_type_assertions(
        ground_truth_object,
        new_point,
        "Cartesian 3d point with named arguments",
    )

    # Geographical point with named arguments
    ground_truth_object = neomodel.contrib.spatial_properties.NeomodelPoint(
        longitude=0.0, latitude=0.0
    )
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(
        longitude=0.0, latitude=0.0
    )
    basic_type_assertions(
        ground_truth_object,
        new_point,
        "Geographical 2d point with named arguments",
    )

    ground_truth_object = neomodel.contrib.spatial_properties.NeomodelPoint(
        longitude=0.0, latitude=0.0, height=0.0
    )
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(
        longitude=0.0, latitude=0.0, height=0.0
    )
    basic_type_assertions(
        ground_truth_object,
        new_point,
        "Geographical 3d point with named arguments",
    )


def test_copy_constructors():
    """
    Tests all the possible ways by which a NeomodelPoint can be instantiated successfully via a copy constructor call.

    :return:
    """

    # Neo4j versions lower than 3.4.0 do not support Point. In that case, skip the test.
    check_and_skip_neo4j_least_version(
        340, "This version does not support spatial data types."
    )

    # Instantiate from Shapely point

    # Implicit cartesian from shapely point
    ground_truth = neomodel.contrib.spatial_properties.NeomodelPoint(
        (0.0, 0.0), crs="cartesian"
    )
    shapely_point = shapely.geometry.Point((0.0, 0.0))
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(shapely_point)
    basic_type_assertions(
        ground_truth, new_point, "Implicit cartesian by shapely Point"
    )

    # Explicit geographical by shapely point
    ground_truth = neomodel.contrib.spatial_properties.NeomodelPoint(
        (0.0, 0.0, 0.0), crs="wgs-84-3d"
    )
    shapely_point = shapely.geometry.Point((0.0, 0.0, 0.0))
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(
        shapely_point, crs="wgs-84-3d"
    )
    basic_type_assertions(
        ground_truth, new_point, "Explicit geographical by shapely Point"
    )

    # Copy constructor for NeomodelPoints
    ground_truth = neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0))
    other_neomodel_point = neomodel.contrib.spatial_properties.NeomodelPoint((0.0, 0.0))
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(other_neomodel_point)
    basic_type_assertions(ground_truth, new_point, "NeomodelPoint copy constructor")


def test_prohibited_constructor_forms():
    """
    Tests all the possible forms by which construction of NeomodelPoints should fail.

    :return:
    """

    # Neo4j versions lower than 3.4.0 do not support Point. In that case, skip the test.
    check_and_skip_neo4j_least_version(
        340, "This version does not support spatial data types."
    )

    # Absurd CRS
    with pytest.raises(ValueError, match=r"Invalid CRS\(blue_hotel\)"):
        _ = neomodel.contrib.spatial_properties.NeomodelPoint((0, 0), crs="blue_hotel")

    # Absurd coord dimensionality
    with pytest.raises(
        ValueError,
    ):
        _ = neomodel.contrib.spatial_properties.NeomodelPoint(
            (0, 0, 0, 0, 0, 0, 0), crs="cartesian"
        )

    # Absurd datatype passed to copy constructor
    with pytest.raises(
        TypeError,
    ):
        _ = neomodel.contrib.spatial_properties.NeomodelPoint(
            "it don't mean a thing if it ain't got that swing",
            crs="cartesian",
        )

    # Trying to instantiate a point with any of BOTH x,y,z or longitude, latitude, height
    with pytest.raises(ValueError, match="Invalid instantiation via arguments"):
        _ = neomodel.contrib.spatial_properties.NeomodelPoint(
            x=0.0,
            y=0.0,
            longitude=0.0,
            latitude=2.0,
            height=-2.0,
            crs="cartesian",
        )

    # Trying to instantiate a point with absolutely NO parameters
    with pytest.raises(ValueError, match="Invalid instantiation via no arguments"):
        _ = neomodel.contrib.spatial_properties.NeomodelPoint()


def test_property_accessors_depending_on_crs_shapely_lt_2():
    """
    Tests that points are accessed via their respective accessors.

    :return:
    """

    # Neo4j versions lower than 3.4.0 do not support Point. In that case, skip the test.
    check_and_skip_neo4j_least_version(
        340, "This version does not support spatial data types."
    )

    # Check the version of Shapely installed to run the appropriate tests:
    try:
        from shapely import __version__
    except ImportError:
        pytest.skip("Shapely not installed")

    if int("".join(__version__.split(".")[0:3])) >= 200:
        pytest.skip("Shapely 2 is installed, skipping earlier version test")

    # Geometrical points only have x,y,z coordinates
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(
        (0.0, 0.0, 0.0), crs="cartesian-3d"
    )
    with pytest.raises(AttributeError, match=r'Invalid coordinate \("longitude"\)'):
        new_point.longitude
    with pytest.raises(AttributeError, match=r'Invalid coordinate \("latitude"\)'):
        new_point.latitude
    with pytest.raises(AttributeError, match=r'Invalid coordinate \("height"\)'):
        new_point.height

    # Geographical points only have longitude, latitude, height coordinates
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(
        (0.0, 0.0, 0.0), crs="wgs-84-3d"
    )
    with pytest.raises(AttributeError, match=r'Invalid coordinate \("x"\)'):
        new_point.x
    with pytest.raises(AttributeError, match=r'Invalid coordinate \("y"\)'):
        new_point.y
    with pytest.raises(AttributeError, match=r'Invalid coordinate \("z"\)'):
        new_point.z


def test_property_accessors_depending_on_crs_shapely_gte_2():
    """
    Tests that points are accessed via their respective accessors.

    :return:
    """

    # Neo4j versions lower than 3.4.0 do not support Point. In that case, skip the test.
    check_and_skip_neo4j_least_version(
        340, "This version does not support spatial data types."
    )

    # Check the version of Shapely installed to run the appropriate tests:
    try:
        from shapely import __version__
    except ImportError:
        pytest.skip("Shapely not installed")

    if int("".join(__version__.split(".")[0:3])) < 200:
        pytest.skip("Shapely < 2.0.0 is installed, skipping test")
    # Geometrical points only have x,y,z coordinates
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(
        (0.0, 0.0, 0.0), crs="cartesian-3d"
    )
    with pytest.raises(TypeError, match=r'Invalid coordinate \("longitude"\)'):
        new_point.longitude
    with pytest.raises(TypeError, match=r'Invalid coordinate \("latitude"\)'):
        new_point.latitude
    with pytest.raises(TypeError, match=r'Invalid coordinate \("height"\)'):
        new_point.height

    # Geographical points only have longitude, latitude, height coordinates
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(
        (0.0, 0.0, 0.0), crs="wgs-84-3d"
    )
    with pytest.raises(TypeError, match=r'Invalid coordinate \("x"\)'):
        new_point.x
    with pytest.raises(TypeError, match=r'Invalid coordinate \("y"\)'):
        new_point.y
    with pytest.raises(TypeError, match=r'Invalid coordinate \("z"\)'):
        new_point.z


def test_property_accessors():
    """
    Tests that points are accessed via their respective accessors and that these accessors return the right values.

    :return:
    """

    # Neo4j versions lower than 3.4.0 do not support Point. In that case, skip the test.
    check_and_skip_neo4j_least_version(
        340, "This version does not support spatial data types."
    )

    # Geometrical points
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(
        (0.0, 1.0, 2.0), crs="cartesian-3d"
    )
    assert new_point.x == 0.0, "Expected x coordinate to be 0.0"
    assert new_point.y == 1.0, "Expected y coordinate to be 1.0"
    assert new_point.z == 2.0, "Expected z coordinate to be 2.0"

    # Geographical points
    new_point = neomodel.contrib.spatial_properties.NeomodelPoint(
        (0.0, 1.0, 2.0), crs="wgs-84-3d"
    )
    assert new_point.longitude == 0.0, "Expected longitude to be 0.0"
    assert new_point.latitude == 1.0, "Expected latitude to be 1.0"
    assert new_point.height == 2.0, "Expected height to be 2.0"



import warnings
from types import FrameType
from typing import Any, Callable, Optional

from neo4j.graph import Entity

OUTGOING, INCOMING, EITHER = 1, -1, 0


def deprecated(message: str) -> Callable:
    # pylint:disable=invalid-name
    def f__(f: Callable) -> Callable:
        def f_(*args, **kwargs) -> Any:  # type: ignore
            warnings.warn(message, category=DeprecationWarning, stacklevel=2)
            return f(*args, **kwargs)

        f_.__name__ = f.__name__
        f_.__doc__ = f.__doc__
        f_.__dict__.update(f.__dict__)
        return f_

    return f__


def classproperty(f: Callable) -> Any:
    class cpf:
        def __init__(self, getter: Callable) -> None:
            self.getter = getter

        def __get__(self, obj: Any, type: Optional[Any] = None) -> Any:
            return self.getter(type)

    return cpf(f)


# Just used for error messages
class _UnsavedNode:
    def __repr__(self) -> str:
        return "<unsaved node>"

    def __str__(self) -> str:
        return self.__repr__()


def get_graph_entity_properties(entity: Entity) -> dict:
    """
    Get the properties from a neo4j.graph.Entity (neo4j.graph.Node or neo4j.graph.Relationship) object.
    """
    return entity._properties


def enumerate_traceback(initial_frame: Optional[FrameType] = None) -> Any:
    depth, frame = 0, initial_frame
    while frame is not None:
        yield depth, frame
        frame = frame.f_back
        depth += 1


def version_tag_to_integer(version_tag: str) -> int:
    """
    Converts a version string to an integer representation to allow for quick comparisons between versions.

    :param a_version_string: The version string to be converted (e.g. '5.4.0')
    :type a_version_string: str
    :return: An integer representation of the version string (e.g. '5.4.0' --> 50400)
    :rtype: int
    """
    components = version_tag.split(".")
    while len(components) < 3:
        components.append("0")
    num = 0
    for index, component in enumerate(components):
        # Aura started adding a -aura suffix in version numbers, like "5.14-aura"
        # This will strip the suffix to allow for proper comparison : 14 instead of 14-aura
        if "-" in component:
            component = component.split("-")[0]
        num += (100 ** ((len(components) - 1) - index)) * int(component)
    return num



from neomodel.exceptions import DeflateConflict, InflateConflict
from neomodel.sync_.core import StructuredNode
from neomodel.util import get_graph_entity_properties


class SemiStructuredNode(StructuredNode):
    """
    A base class allowing properties to be stored on a node that aren't
    specified in its definition. Conflicting properties are signaled with the
    :class:`DeflateConflict` exception::

        class Person(SemiStructuredNode):
            name = StringProperty()
            age = IntegerProperty()

            def hello(self):
                print("Hi my names " + self.name)

        tim = Person(name='Tim', age=8, weight=11).save()
        tim.hello = "Hi"
        tim.save() # DeflateConflict
    """

    __abstract_node__ = True

    @classmethod
    def inflate(cls, node):
        # Inflate all properties registered in the class definition
        snode = super().inflate(node)

        # Node can be a string or int for lazy loading (See StructuredNode.inflate). In that case, `node` has nothing
        # that can be unpacked further.
        if not hasattr(node, "items"):
            return snode

        # Inflate all extra properties not registered in the class definition
        registered_db_property_names = {
            property.get_db_property_name(name)
            for name, property in cls.defined_properties(
                aliases=False, rels=False
            ).items()
        }
        extra_keys = node.keys() - registered_db_property_names
        for extra_key in extra_keys:
            value = node[extra_key]
            if hasattr(cls, extra_key):
                raise InflateConflict(cls, extra_key, value, snode.element_id)
            setattr(snode, extra_key, value)

        return snode

    @classmethod
    def deflate(cls, node_props, obj=None, skip_empty=False):
        # Deflate all properties registered in the class definition
        deflated = super().deflate(node_props, obj, skip_empty=skip_empty)

        # Deflate all extra properties not registered in the class definition
        registered_names = cls.defined_properties(aliases=False, rels=False).keys()
        extra_keys = node_props.keys() - registered_names
        for extra_key in extra_keys:
            value = node_props[extra_key]
            if hasattr(cls, extra_key):
                raise DeflateConflict(
                    cls, extra_key, value, node_props.get("element_id")
                )
            deflated[extra_key] = node_props[extra_key]

        return deflated



import json
import subprocess

import pytest

from neomodel import (
    RelationshipTo,
    StringProperty,
    StructuredNode,
    StructuredRel,
    config,
)
from neomodel.sync_.core import db


class ScriptsTestRel(StructuredRel):
    some_unique_property = StringProperty(unique_index=db.version_is_higher_than("5.7"))
    some_index_property = StringProperty(index=True)


class ScriptsTestNode(StructuredNode):
    personal_id = StringProperty(unique_index=True)
    name = StringProperty(index=True)
    rel = RelationshipTo("ScriptsTestNode", "REL", model=ScriptsTestRel)
    other_rel = RelationshipTo("ScriptsTestNode", "OTHER_REL")


def test_neomodel_install_labels():
    result = subprocess.run(
        ["neomodel_install_labels", "--help"],
        capture_output=True,
        text=True,
        check=False,
    )
    assert "usage: neomodel_install_labels" in result.stdout
    assert result.returncode == 0

    result = subprocess.run(
        ["neomodel_install_labels", "test/test_scripts.py", "--db", db.url],
        capture_output=True,
        text=True,
        check=False,
    )
    assert result.returncode == 0
    assert "Setting up indexes and constraints" in result.stdout
    constraints = db.list_constraints()
    parsed_constraints = [
        (element["type"], element["labelsOrTypes"], element["properties"])
        for element in constraints
    ]
    assert ("UNIQUENESS", ["ScriptsTestNode"], ["personal_id"]) in parsed_constraints
    if db.version_is_higher_than("5.7"):
        assert (
            "RELATIONSHIP_UNIQUENESS",
            ["REL"],
            ["some_unique_property"],
        ) in parsed_constraints
    indexes = db.list_indexes()
    parsed_indexes = [
        (element["labelsOrTypes"], element["properties"]) for element in indexes
    ]
    assert (["ScriptsTestNode"], ["name"]) in parsed_indexes
    assert (["REL"], ["some_index_property"]) in parsed_indexes


def test_neomodel_remove_labels():
    result = subprocess.run(
        ["neomodel_remove_labels", "--help"],
        capture_output=True,
        text=True,
        check=False,
    )
    assert "usage: neomodel_remove_labels" in result.stdout
    assert result.returncode == 0

    result = subprocess.run(
        ["neomodel_remove_labels", "--db", config.DATABASE_URL],
        capture_output=True,
        text=True,
        check=False,
    )
    assert (
        "Dropping unique constraint and index on label ScriptsTestNode" in result.stdout
    )
    assert result.returncode == 0
    constraints = db.list_constraints()
    indexes = db.list_indexes(exclude_token_lookup=True)
    assert len(constraints) == 0
    assert len(indexes) == 0


@pytest.mark.parametrize(
    "script_flavour",
    [
        "",
        "_light",
    ],
)
def test_neomodel_inspect_database(script_flavour):
    output_file = "test/data/neomodel_inspect_database_test_output.py"
    # Check that the help option works
    result = subprocess.run(
        ["neomodel_inspect_database", "--help"],
        capture_output=True,
        text=True,
        check=False,
    )
    assert "usage: neomodel_inspect_database" in result.stdout
    assert result.returncode == 0

    db.clear_neo4j_database()
    db.install_labels(ScriptsTestNode)
    db.install_labels(ScriptsTestRel)

    # Create a few nodes and a rel, with indexes and constraints
    node1 = ScriptsTestNode(personal_id="1", name="test").save()
    node2 = ScriptsTestNode(personal_id="2", name="test").save()
    node3 = ScriptsTestNode(personal_id="3", name="test").save()
    node1.rel.connect(node2, {"some_unique_property": "1", "some_index_property": "2"})
    node1.other_rel.connect(node3)

    # Create a node with all the parsable property types
    # Also create a node with no properties
    db.cypher_query(
        """
        CREATE (:EveryPropertyTypeNode {
            string_property: "Hello World",
            boolean_property: true,
            date_property: date("2020-01-01"),
            datetime_property: datetime("2020-01-01T00:00:00.000Z"),
            integer_property: 1,
            float_property: 1.0,
            point_property: point({x: 0.0, y: 0.0, crs: "wgs-84"}),
            array_property: ["test"]
        })
        CREATE (:NoPropertyNode)
        CREATE (n1:NoPropertyRelNode)
        CREATE (n2:NoPropertyRelNode)
        CREATE (n1)-[:NO_PROP_REL]->(n2)
        """
    )

    # Test the console output version of the script
    args_list = ["neomodel_inspect_database", "--db", config.DATABASE_URL]
    if script_flavour == "_light":
        args_list += ["--no-rel-props", "--no-rel-cardinality"]
    result = subprocess.run(
        args_list,
        capture_output=True,
        text=True,
        check=True,
    )

    wrapped_console_output = [
        line for line in result.stdout.splitlines() if line.strip()
    ]
    assert wrapped_console_output[0].startswith("Connecting to")
    # Check that all the expected lines are here
    file_path = (
        f"test/data/neomodel_inspect_database_output{script_flavour}.txt"
        if db.version_is_higher_than("5.7")
        else f"test/data/neomodel_inspect_database_output_pre_5_7{script_flavour}.txt"
    )
    with open(file_path, "r") as f:
        wrapped_test_file = [line for line in f.read().split("\n") if line.strip()]
        for line in wrapped_test_file:
            # The neomodel components import order might differ
            # So let's check that every import that should be added is added, regardless of order
            if line.startswith("from neomodel import"):
                parsed_imports = line.replace("from neomodel import ", "").split(", ")
                expected_imports = (
                    wrapped_console_output[1]
                    .replace("from neomodel import ", "")
                    .split(", ")
                )
                assert set(parsed_imports) == set(expected_imports)
                wrapped_test_file.remove(line)
                break

        # Check that both outputs have the same set of lines, regardless of order and redundance
        assert set(wrapped_test_file) == set(wrapped_console_output[2:])

    # Test the file output version of the script
    args_list += ["--write-to", output_file]
    result = subprocess.run(
        args_list,
        capture_output=True,
        text=True,
        check=True,
    )

    # Check that the file was written
    # And that the file output has the same content as the console output
    # Again, regardless of order and redundance
    wrapped_file_console_output = [
        line for line in result.stdout.splitlines() if line.strip()
    ]
    assert wrapped_file_console_output[0].startswith("Connecting to")
    assert wrapped_file_console_output[1].startswith("Writing to")
    with open(output_file, "r") as f:
        wrapped_output_file = [line for line in f.read().split("\n") if line.strip()]
        assert set(wrapped_output_file) == set(wrapped_console_output[1:])

    # Finally, delete the file created by the script
    subprocess.run(
        ["rm", output_file],
    )


def test_neomodel_generate_diagram():
    result = subprocess.run(
        ["neomodel_generate_diagram", "--help"],
        capture_output=True,
        text=True,
        check=False,
    )
    assert "usage: neomodel_generate_diagram" in result.stdout
    assert result.returncode == 0

    output_dir = "test/data"

    # Arrows
    result = subprocess.run(
        [
            "neomodel_generate_diagram",
            "test/diagram_classes.py",
            "--file-type",
            "arrows",
            "--write-to-dir",
            output_dir,
        ],
        capture_output=True,
        text=True,
        check=False,
    )
    assert "Loaded test/diagram_classes.py" in result.stdout
    assert "Successfully wrote diagram to file" in result.stdout
    assert result.returncode == 0

    # Check that the output file is as expected
    with open("test/data/model_diagram.json", "r", encoding="utf-8") as f:
        actual_json = json.loads(f.read())
    with open("test/data/expected_model_diagram.json", "r", encoding="utf-8") as f:
        expected_json = json.loads(f.read())
    assert actual_json["style"] == expected_json["style"]
    assert len(actual_json["nodes"]) == len(expected_json["nodes"])
    assert len(actual_json["relationships"]) == len(expected_json["relationships"])

    for index, node in enumerate(actual_json["nodes"]):
        expected_node = expected_json["nodes"][index]
        assert node["id"] == expected_node["id"]
        assert node["labels"] == expected_node["labels"]
        assert node["properties"] == expected_node["properties"]

    assert actual_json["relationships"] == expected_json["relationships"]

    # PlantUML
    puml_result = subprocess.run(
        [
            "neomodel_generate_diagram",
            "test/diagram_classes.py",
            "--file-type",
            "puml",
            "--write-to-dir",
            output_dir,
        ],
        capture_output=True,
        text=True,
        check=False,
    )
    assert "Loaded test/diagram_classes.py" in result.stdout
    assert "Successfully wrote diagram to file" in result.stdout
    assert puml_result.returncode == 0

    # Check that the output file is as expected
    with open("test/data/model_diagram.puml", "r", encoding="utf-8") as f:
        actual_json = f.read()
    with open("test/data/expected_model_diagram.puml", "r", encoding="utf-8") as f:
        expected_json = f.read()
    assert actual_json == expected_json

    # Wrong format
    wrong_result = subprocess.run(
        [
            "neomodel_generate_diagram",
            "test/diagram_classes.py",
            "--file-type",
            "pdf",
            "--write-to-dir",
            output_dir,
        ],
        capture_output=True,
        text=True,
        check=False,
    )
    assert "Unsupported file type : pdf" in wrong_result.stderr
